__all__ = (
    "CharClass",
    "Artificer",
    "Barbarian",
    "Bard",
    "Cleric",
    "Druid",
    "Fighter",
    "Monk",
    "Paladin",
    "Ranger",
    "Rogue",
    "Sorceror",
    "Warlock",
    "Wizard",
    "RevisedRanger",
    "available_classes",
)

from dungeonsheets.classes.artificer import Artificer
from dungeonsheets.classes.barbarian import Barbarian
from dungeonsheets.classes.bard import Bard
from dungeonsheets.classes.classes import CharClass
from dungeonsheets.classes.cleric import Cleric
from dungeonsheets.classes.druid import Druid
from dungeonsheets.classes.fighter import Fighter
from dungeonsheets.classes.monk import Monk
from dungeonsheets.classes.paladin import Paladin
from dungeonsheets.classes.ranger import Ranger, RevisedRanger
from dungeonsheets.classes.rogue import Rogue
from dungeonsheets.classes.sorceror import Sorceror
from dungeonsheets.classes.warlock import Warlock
from dungeonsheets.classes.wizard import Wizard

available_classes = [
    Artificer,
    Barbarian,
    Bard,
    Cleric,
    Druid,
    Fighter,
    Monk,
    Paladin,
    Ranger,
    Rogue,
    Sorceror,
    Warlock,
    Wizard,
    RevisedRanger,
]



import math
from collections import namedtuple
from math import ceil
import logging

from dungeonsheets.armor import HeavyArmor, NoArmor, NoShield
from dungeonsheets.features import (
    AmbushMaster,
    Defense,
    DraconicResilience,
    DreadAmbusher,
    FastMovement,
    FeralInstinct,
    GiftOfTheDepths,
    JackOfAllTrades,
    NaturalArmor,
    NaturalExplorerRevised,
    QuickDraw,
    RakishAudacity,
    RemarkableAthlete,
    SeaSoul,
    SoulOfTheForge,
    SuperiorMobility,
    UnarmoredDefenseBarbarian,
    UnarmoredDefenseMonk,
    UnarmoredMovement,
)


log = logging.getLogger(__name__)


def mod_str(modifier):
    """Converts a modifier to a string, eg 2 -> '+2'."""
    return "{:+d}".format(modifier)


def ability_mod_str(character, ability):
    return mod_str(getattr(character, ability).modifier)


def stat_abbreviation(stat_name):
    """Abbreviate the name of an ability."""
    return stat_name.upper()[:3]


AbilityScore = namedtuple("AbilityScore", ("value", "modifier", "saving_throw", "name"))


class Ability:
    ability_name = None

    def __init__(self, default_value=10):
        self.default_value = default_value

    def __set_name__(self, character, name):
        self.ability_name = name

    def _check_dict(self, obj):
        if not hasattr(obj, "_ability_scores"):
            # No ability score dictionary exists
            obj._ability_scores = {self.ability_name: self.default_value}
        elif self.ability_name not in obj._ability_scores.keys():
            # ability score dictionary exists but doesn't have this ability
            obj._ability_scores[self.ability_name] = self.default_value

    def __get__(self, actor, Actor):
        self._check_dict(actor)
        score = actor._ability_scores[self.ability_name]
        modifier = math.floor((score - 10) / 2)
        # Check for proficiency
        saving_throw = modifier
        if self.ability_name is not None and hasattr(
            actor, "saving_throw_proficiencies"
        ):
            is_proficient = self.ability_name in actor.saving_throw_proficiencies
            if is_proficient:
                saving_throw += actor.proficiency_bonus
        # Check for bonuses to saving throws from magic items
        for mitem in actor.magic_items:
            saving_throw += mitem.st_bonus(ability=self.ability_name)
            # saving_throw += getattr(mitem, "st_bonus", 0)
        # Create the named tuple
        value = AbilityScore(modifier=modifier, value=score, saving_throw=saving_throw, name=self.ability_name)
        return value

    def __set__(self, actor, val):
        self._check_dict(actor)
        actor._ability_scores[self.ability_name] = val
        self.value = val


class Skill:
    """An ability-based skill, such as athletics.

    Attributes
    ----------
    ability_name:
      The name of the ability, as a python-compatible string.
    skill_name:
      The name of the base ability that determines the skill
      modifier.
    is_proficient
      Bool that describes if the owner is proficient in this skill.
    is_expertise
      Bool that describes if the owner has expertise in this skill.
    is_jack_of_all_trades
      Bool that describes if this skill benefits from Jack of All
      Trades feature (False if already proficient).
    is_remarkable_athlete
      Bool that describes if this skill benefits from Remarkable
      Athlete feature (False if already proficient).
    modifier
      The base ability modifier, after relevant proficiency bonuses
      have been applied.
    proficiency_modifier
      The bonus that is applied to the base ability. Usually the same
      as the owner's proficiency bonus if ``is_proficient`` is True,
      but can be different based on class features.,

    """
    
    ability_name = ""
    skill_name = ""
    actor = None
    
    def __init__(self, ability):
        self.ability_name = ability

    def __set_name__(self, owner, name):
        self.skill_name = name.lower().replace("_", " ")

    def __get__(self, actor, owner):
        self.actor = actor
        return self

    def __str__(self):
        return self.skill_name.title()

    @property
    def is_remarkable_athlete(self):
        already_proficient = (self.is_proficient or self.is_expertise)
        if self.actor.has_feature(RemarkableAthlete) and not already_proficient:
            return True
        else:
            return False

    @property
    def is_jack_of_all_trades(self):
        already_proficient = (self.is_proficient or self.is_expertise or self.is_remarkable_athlete)
        if self.actor.has_feature(JackOfAllTrades) and not already_proficient:
            return True
        else:
            return False

    @property
    def is_proficient(self):
        # Check for proficiency
        proficiencies = [p.replace("_", " ") for p in self.actor.skill_proficiencies]
        is_proficient = self.skill_name in proficiencies
        return is_proficient

    @property
    def is_expertise(self):
        return self.skill_name in self.actor.skill_expertise

    @property
    def proficiency_modifier(self):
        modifier = 0
        if self.is_proficient:
            modifier += self.actor.proficiency_bonus
        if self.is_remarkable_athlete:
            modifier += ceil(self.actor.proficiency_bonus / 2.0)
        if self.is_jack_of_all_trades:
            modifier += self.actor.proficiency_bonus // 2
        # Check for expertise
        if self.is_expertise:
            modifier += self.actor.proficiency_bonus
        return modifier

    @property
    def modifier(self):
        ability = getattr(self.actor, self.ability_name)
        modifier = ability.modifier + self.proficiency_modifier
        log.info("%s modifier for '%s': %d", self, self.actor.name, modifier)
        return modifier


class ArmorClass:
    """
    The Armor Class of a character
    """

    def __get__(self, actor, Actor):
        armor = actor.armor or NoArmor()
        ac = armor.base_armor_class
        # calculate and apply modifiers
        if armor.dexterity_applied:
            if armor.dexterity_mod_max is None:
                ac += actor.dexterity.modifier
            else:
                ac += min(actor.dexterity.modifier, armor.dexterity_mod_max)
        if actor.has_feature(NaturalArmor):
            ac = max(ac, 13 + actor.dexterity.modifier)
        shield = actor.shield or NoShield()
        ac += shield.base_armor_class
        # Compute feature-specific additions
        if actor.has_feature(UnarmoredDefenseMonk):
            if isinstance(armor, NoArmor) and isinstance(shield, NoShield):
                ac += actor.wisdom.modifier
        if actor.has_feature(UnarmoredDefenseBarbarian):
            if isinstance(armor, NoArmor):
                ac += actor.constitution.modifier
        if actor.has_feature(DraconicResilience):
            if isinstance(armor, NoArmor):
                ac += 3
        if actor.has_feature(Defense):
            if not isinstance(armor, NoArmor):
                ac += 1
        if actor.has_feature(SoulOfTheForge):
            if isinstance(armor, HeavyArmor):
                ac += 1
        # Check if any magic items add to AC
        for mitem in actor.magic_items:
            if hasattr(mitem, "ac_bonus"):
                ac += mitem.ac_bonus
        return ac


class Speed:
    """
    The speed of a character
    """

    def __get__(self, actor, Actor):
        speed = actor.race.speed
        other_speed = ""
        if isinstance(speed, str):
            other_speed = speed[2:]
            speed = int(speed[:2])  # ignore other speeds, like fly
        if actor.has_feature(FastMovement):
            if not isinstance(actor.armor, HeavyArmor):
                speed += 10
        if actor.has_feature(SuperiorMobility):
            speed += 10
        if isinstance(actor.armor, NoArmor) or (actor.armor is None):
            for f in actor.features:
                if isinstance(f, UnarmoredMovement):
                    speed += f.speed_bonus
        if actor.has_feature(GiftOfTheDepths):
            if "swim" not in other_speed:
                other_speed += " ({:d} swim)".format(speed)
        if actor.has_feature(SeaSoul):
            if "swim" not in other_speed:
                other_speed += " (30 swim)"
        return "{:d}{:s}".format(speed, other_speed)


class NumericalInitiative:
    """A numerical representation of initiative"""

    def __get__(self, actor, Actor):
        ini = actor.dexterity.modifier
        if actor.has_feature(QuickDraw):
            ini += actor.proficiency_bonus
        if actor.has_feature(DreadAmbusher):
            ini += actor.wisdom.modifier
        if actor.has_feature(RakishAudacity):
            ini += actor.charisma.modifier

        has_advantage = (
            actor.has_feature(NaturalExplorerRevised)
            or actor.has_feature(FeralInstinct)
            or actor.has_feature(AmbushMaster)
        )
        return ini, has_advantage


class Initiative(NumericalInitiative):
    """A character's initiative"""

    def __get__(self, actor, Actor):
        ini, has_advantage = super(Initiative, self).__get__(actor, Actor)
        ini = "{:+d}".format(ini)
        if has_advantage:
            ini += "(A)"
        return ini



from dungeonsheets import features as feats
from dungeonsheets.content_registry import default_content_registry


default_content_registry.add_module(__name__)


class Background:
    name = "Generic background"
    owner = None
    skill_proficiencies = ()
    weapon_proficiencies = ()
    proficiencies_text = ()
    skill_choices = ()
    num_skill_choices = 0
    features = ()
    languages = ()

    def __init__(self, owner=None):
        self.owner = owner
        cls = type(self)
        self.features = tuple([f(owner=self.owner) for f in cls.features])

    def __str__(self):
        return self.name


class Acolyte(Background):
    name = "Acolyte"
    skill_proficiencies = ("insight", "religion")
    languages = ("[choose one]", "[choose one]")
    features = (feats.ShelterOfTheFaithful,)


class Charlatan(Background):
    name = "Charlatan"
    skill_proficiencies = ("deception", "sleight of hand")
    features = (feats.FalseIdentity,)


class Criminal(Background):
    name = "Criminal"
    skill_proficiencies = ("deception", "stealth")
    features = (feats.CriminalContact,)


class Spy(Criminal):
    name = "Spy"


class Entertainer(Background):
    name = "Entertainer"
    skill_proficiencies = ("acrobatics", "performance")
    features = (feats.ByPopularDemand,)


class Gladiator(Entertainer):
    name = "Gladiator"


class FolkHero(Background):
    name = "Folk Hero"
    skill_proficiencies = ("animal handling", "survival")
    features = (feats.RusticHospitality,)


class GuildArtisan(Background):
    name = "Guild Artisan"
    skill_proficiencies = ("insight", "persuasion")
    languages = ("[choose one]", "[choose one]")
    features = (feats.GuildMembership,)


class GuildMerchant(GuildArtisan):
    name = "Guild Merchant"


class Hermit(Background):
    name = "Hermit"
    skill_proficiencies = ("medicine", "religion")
    languages = ("[choose one]",)
    features = (feats.Discovery,)


class Noble(Background):
    name = "Noble"
    skill_proficiencies = ("history", "persuasion")
    languages = ("[choose one]",)
    features = (feats.PositionOfPrivilege,)


class Knight(Noble):
    name = "Knight"


class Outlander(Background):
    name = "Outlander"
    skill_proficiencies = ("athletics", "survival")
    languages = ("[choose one]",)
    features = (feats.Wanderer,)


class RivalIntern(Background):
    """You were an intern at a rival of Acquisitions Incorporated, and you
    gained a healthy respect for nocjusc the job and the franchising
    opportunities. but for the ruth- less and efficient way
    Acquisitions Incorporated goes about its business. Why deal with
    the rest, when you can work for the best?

    Perhaps the rival did not treat you as well as you were hoping, or
    you washed out of that organization. Maybe you hope to leverage
    the knowledge you gained there for an advantage at Acquisitions
    Incorporated. Either way, you're now bringing your talents to the
    company, ready to put your skills lo use.

    """

    name = "Rival Intern"
    skill_proficiencies = ("history", "investigation")
    proficiencies_text = ("One type of artisan's tools",)
    languages = ("[choose one]",)
    features = (feats.InsideInformant,)


class Sage(Background):
    name = "Sage"
    skill_proficiencies = ("arcana", "history")
    languages = ("[choose one]", "[choose one]")
    features = (feats.Researcher,)


class Sailor(Background):
    name = "Sailor"
    skill_proficiencies = ("athletics", "perception")
    features = (feats.ShipsPassage,)


class Pirate(Sailor):
    name = "Pirate"


class Soldier(Background):
    name = "Soldier"
    skill_proficiencies = ("athletics", "intimidation")
    features = (feats.MilitaryRank,)


class Urchin(Background):
    name = "Urchin"
    skill_proficiencies = ("sleight of hand", "stealth")
    features = (feats.CitySecrets,)


# Sword's Coast Adventurers Guide
class CityWatch(Background):
    name = "City Watch"
    skill_proficiencies = ("athletics", "insight")
    languages = ("[choose one]", "[choose one]")
    features = (feats.WatchersEye,)


class ClanCrafter(Background):
    name = "Clan Crafter"
    skill_proficiencies = ("history", "insight")
    languages = ("Dwarvish",)
    features = (feats.RespectOfTheStoutFolk,)


class CloisteredScholar(Background):
    name = "Cloistered Scholar"
    skill_proficiencies = ("history",)
    skill_choices = ("arcana", "nature", "religion")
    num_skill_choices = 1
    languages = ("[choose one]", "[choose one]")
    features = (feats.LibraryAccess,)


class Courtier(Background):
    name = "Courtier"
    skill_proficiencies = ("insight", "persuasion")
    languages = ("[choose one]", "[choose one]")
    features = (feats.CourtFunctionary,)


class FactionAgent(Background):
    name = "Faction Agent"
    skill_proficiencies = ("insight",)
    skill_choices = (
        "animal handling",
        "arcana",
        "deception",
        "history",
        "intimidation",
        "investigation",
        "medicine",
        "nature",
        "perception",
        "performance",
        "persuasion",
        "religion",
        "survival",
    )
    num_skill_choices = 1
    languages = ("[choose one]", "[choose one]")
    features = (feats.SafeHaven,)


class FarTraveler(Background):
    name = "Far Traveler"
    skill_proficiencies = ("insight", "perception")
    languages = ("[choose one]",)
    features = (feats.AllEyesOnYou,)


class Inheritor(Background):
    name = "Inheritor"
    skill_proficiencies = ("survival",)
    skill_choices = ("arcana", "history", "religion")
    num_skill_choices = 1
    languages = ("[choose one]",)
    features = (feats.Inheritance,)


class KnightOfTheOrder(Background):
    name = "Knight of the Order"
    skill_proficiencies = ("persuasion",)
    skill_choices = ("arcana", "history", "nature", "religion")
    num_skill_choices = 1
    languages = ("[choose one]",)
    features = (feats.KnightlyRegard,)


class MercenaryVeteran(Background):
    name = "Mercenary Veteran"
    skill_proficiencies = ("athletics", "persuasion")
    features = (feats.MercenaryLife,)


class UrbanBountyHunter(Background):
    name = "Urban Bounty Hunter"
    skill_proficiencies = ()
    skill_choices = ("Deception", "Insight", "Persuasion", "Stealth")
    num_skill_choices = 2
    features = (feats.EarToTheGround,)


class UthgardtTribeMember(Background):
    name = "Uthgardt Tribe Member"
    skill_profifiencies = ("athletics", "survival")
    languages = ("[choose one]",)
    features = (feats.UthgardtHeritage,)


class WaterdhavianNoble(Background):
    name = "Waterdhavian Noble"
    skill_proficiencies = ("history", "persuasion")
    languages = ("[choose one]",)
    features = (feats.KeptInStyle,)


PHB_backgrounds = [
    Acolyte,
    Charlatan,
    Criminal,
    Spy,
    Entertainer,
    Gladiator,
    FolkHero,
    GuildArtisan,
    GuildMerchant,
    Hermit,
    Noble,
    Knight,
    Outlander,
    Sage,
    Sailor,
    Pirate,
    Soldier,
    Urchin,
]

SCAG_backgrounds = [
    CityWatch,
    ClanCrafter,
    CloisteredScholar,
    Courtier,
    FactionAgent,
    FarTraveler,
    Inheritor,
    KnightOfTheOrder,
    MercenaryVeteran,
    UrbanBountyHunter,
    UthgardtTribeMember,
    WaterdhavianNoble,
]

available_backgrounds = PHB_backgrounds + SCAG_backgrounds

__all__ = tuple([b.name for b in available_backgrounds]) + (
    "PHB_backgrounds",
    "SCAG_backgrounds",
    "available_backgrounds",
)



"""
A collection of monsters from D&D 5e's SRD

This file was autogenerated by https://github.com/stravajiaxen/5e-srd-to-py
"""

from dungeonsheets.monsters.monsters import Monster
from dungeonsheets.stats import Ability


class Rakshasa(Monster):
    """
    Limited Magic Immunity.
      The rakshasa can't be affected or detected by spells of 6th level or
      lower unless it wishes to be. It has advantage on saving throws
      against all other spells and magical effects.
    Innate Spellcasting.
      The rakshasa's innate spellcasting ability is Charisma (spell save DC
      18, +10 to hit with spell attacks). The rakshasa can innately cast the
      following spells, requiring no material components:
      
  
      
      At will: detect thoughts, disguise self, mage hand, minor illusion
      
      3/day each: charm person, detect magic, invisibility, major image,
      suggestion
      
      1/day each: dominate person, fly, plane shift, true seeing
    Multiattack.
      The rakshasa makes two claw attacks
    Claw.
      Melee Weapon Attack: +7 to hit, reach 5 ft., one target. Hit: 9 (2d6 +
      2) slashing damage, and the target is cursed if it is a creature. The
      magical curse takes effect whenever the target takes a short or long
      rest, filling the target's thoughts with horrible images and dreams.
      The cursed target gains no benefit from finishing a short or long
      rest. The curse lasts until it is lifted by a remove curse spell or
      similar magic.
    """
    name = "Rakshasa"
    description = "Medium fiend, lawful evil"
    challenge_rating = 13
    armor_class = 16
    skills = "Deception +10, Insight +8"
    senses = "Darkvision 60 ft., Passive Perception 13"
    languages = "Common, Infernal"
    strength = Ability(14)
    dexterity = Ability(17)
    constitution = Ability(18)
    intelligence = Ability(13)
    wisdom = Ability(16)
    charisma = Ability(20)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 110
    hit_dice = "13d8"
    spells = []


class Rat(Monster):
    """
    Keen Smell.
      The rat has advantage on Wisdom (Perception) checks that rely on
      smell.
    Bite.
      Melee Weapon Attack: +0 to hit, reach 5 ft., one target. Hit: 1
      piercing damage.
    """
    name = "Rat"
    description = "Tiny beast, unaligned"
    challenge_rating = 0
    armor_class = 10
    skills = ""
    senses = "Darkvision 30 ft., Passive Perception 10"
    languages = ""
    strength = Ability(2)
    dexterity = Ability(11)
    constitution = Ability(9)
    intelligence = Ability(2)
    wisdom = Ability(10)
    charisma = Ability(4)
    speed = 20
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 1
    hit_dice = "1d4"
    spells = []


class Raven(Monster):
    """
    Mimicry.
      The raven can mimic simple sounds it has heard, such as a person
      whispering, a baby crying, or an animal chittering. A creature that
      hears the sounds can tell they are imitations with a successful DC 10
      Wisdom (Insight) check.
    Beak.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one target. Hit: 1
      piercing damage.
    """
    name = "Raven"
    description = "Tiny beast, unaligned"
    challenge_rating = 0
    armor_class = 12
    skills = "Perception +3"
    senses = "Passive Perception 13"
    languages = ""
    strength = Ability(2)
    dexterity = Ability(14)
    constitution = Ability(8)
    intelligence = Ability(2)
    wisdom = Ability(12)
    charisma = Ability(6)
    speed = 10
    swim_speed = 0
    fly_speed = 50
    climb_speed = 0
    hp_max = 1
    hit_dice = "1d4"
    spells = []


class RedDragonWyrmling(Monster):
    """
    Bite.
      Melee Weapon Attack: +6 to hit, reach 5 ft., one target. Hit: 9 (1d10
      + 4) piercing damage plus 3 (1d6) fire damage.
    Fire Breath.
      The dragon exhales fire in a 15-foot cone. Each creature in that area
      must make a DC l3 Dexterity saving throw, taking 24 (7d6) fire damage
      on a failed save, or half as much damage on a successful one.
    """
    name = "Red Dragon Wyrmling"
    description = "Medium dragon, chaotic evil"
    challenge_rating = 4
    armor_class = 17
    skills = "Perception +4, Stealth +2"
    senses = "Blindsight 10 ft., Darkvision 60 ft., Passive Perception 14"
    languages = "Draconic"
    strength = Ability(19)
    dexterity = Ability(10)
    constitution = Ability(17)
    intelligence = Ability(12)
    wisdom = Ability(11)
    charisma = Ability(15)
    speed = 30
    swim_speed = 0
    fly_speed = 60
    climb_speed = 30
    hp_max = 75
    hit_dice = "10d8"
    spells = []


class ReefShark(Monster):
    """
    Pack Tactics.
      The shark has advantage on an attack roll against a creature if at
      least one of the shark's allies is within 5 ft. of the creature and
      the ally isn't incapacitated.
    Water Breathing.
      The shark can breathe only underwater.
    Bite.
      Melee Weapon Attack: +4 to hit, reach 5 ft., one target. Hit: 6 (1d8 +
      2) piercing damage.
    """
    name = "Reef Shark"
    description = "Medium beast, unaligned"
    challenge_rating = 0.5
    armor_class = 12
    skills = "Perception +2"
    senses = "Blindsight 30 ft., Passive Perception 12"
    languages = ""
    strength = Ability(14)
    dexterity = Ability(13)
    constitution = Ability(13)
    intelligence = Ability(1)
    wisdom = Ability(10)
    charisma = Ability(4)
    speed = 0
    swim_speed = 40
    fly_speed = 0
    climb_speed = 0
    hp_max = 22
    hit_dice = "4d8"
    spells = []


class Remorhaz(Monster):
    """
    Heated Body.
      A creature that touches the remorhaz or hits it with a melee attack
      while within 5 feet of it takes 10 (3d6) fire damage.
    Bite.
      Melee Weapon Attack: +11 to hit, reach 10 ft., one target. Hit: 40
      (6d10 + 7) piercing damage plus 10 (3d6) fire damage. If the target is
      a creature, it is grappled (escape DC 17). Until this grapple ends,
      the target is restrained, and the remorhaz can't bite another target.
    Swallow.
      The remorhaz makes one bite attack against a Medium or smaller
      creature it is grappling. If the attack hits, that creature takes the
      bite's damage and is swallowed, and the grapple ends. While swallowed,
      the creature is blinded and restrained, it has total cover against
      attacks and other effects outside the remorhaz, and it takes 21 (6d6)
      acid damage at the start of each of the remorhaz's turns.
      
      If the remorhaz takes 30 damage or more on a single turn from a
      creature inside it, the remorhaz must succeed on a DC 15 Constitution
      saving throw at the end of that turn or regurgitate all swallowed
      creatures, which fall prone in a space within 10 feet oft he remorhaz.
      If the remorhaz dies, a swallowed creature is no longer restrained by
      it and can escape from the corpse using 15 feet of movement, exiting
      prone.
    """
    name = "Remorhaz"
    description = "Huge monstrosity, unaligned"
    challenge_rating = 11
    armor_class = 17
    skills = ""
    senses = "Darkvision 60 ft., Tremorsense 60 ft., Passive Perception 10"
    languages = ""
    strength = Ability(24)
    dexterity = Ability(13)
    constitution = Ability(21)
    intelligence = Ability(4)
    wisdom = Ability(10)
    charisma = Ability(5)
    speed = 30
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 195
    hit_dice = "17d12"
    spells = []


class Rhinoceros(Monster):
    """
    Charge.
      If the rhinoceros moves at least 20 ft. straight toward a target and
      then hits it with a gore attack on the same turn, the target takes an
      extra 9 (2d8) bludgeoning damage. If the target is a creature, it must
      succeed on a DC 15 Strength saving throw or be knocked prone.
    Gore.
      Melee Weapon Attack: +7 to hit, reach 5 ft., one target. Hit: 14 (2d8
      + 5) bludgeoning damage.
    """
    name = "Rhinoceros"
    description = "Large beast, unaligned"
    challenge_rating = 2
    armor_class = 11
    skills = ""
    senses = "Passive Perception 11"
    languages = ""
    strength = Ability(21)
    dexterity = Ability(8)
    constitution = Ability(15)
    intelligence = Ability(2)
    wisdom = Ability(12)
    charisma = Ability(6)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 45
    hit_dice = "6d10"
    spells = []


class RidingHorse(Monster):
    """
    Hooves.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one target. Hit: 8 (2d4 +
      3) bludgeoning damage.
    """
    name = "Riding Horse"
    description = "Large beast, unaligned"
    challenge_rating = 0.25
    armor_class = 10
    skills = ""
    senses = "Passive Perception 10"
    languages = ""
    strength = Ability(16)
    dexterity = Ability(10)
    constitution = Ability(12)
    intelligence = Ability(2)
    wisdom = Ability(11)
    charisma = Ability(7)
    speed = 60
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 13
    hit_dice = "2d10"
    spells = []


class Roc(Monster):
    """
    Keen Sight.
      The roc has advantage on Wisdom (Perception) checks that rely on
      sight.
    Multiattack.
      The roc makes two attacks: one with its beak and one with its talons.
    Beak.
      Melee Weapon Attack: +13 to hit, reach 10 ft., one target. Hit: 27
      (4d8 + 9) piercing damage.
    Talons.
      Melee Weapon Attack: +13 to hit, reach 5 ft., one target. Hit: 23 (4d6
      + 9) slashing damage, and the target is grappled (escape DC 19). Until
      this grapple ends, the target is restrained, and the roc can't use its
      talons on another target.
    """
    name = "Roc"
    description = "Gargantuan monstrosity, unaligned"
    challenge_rating = 11
    armor_class = 15
    skills = "Perception +4"
    senses = "Passive Perception 14"
    languages = ""
    strength = Ability(28)
    dexterity = Ability(10)
    constitution = Ability(20)
    intelligence = Ability(3)
    wisdom = Ability(10)
    charisma = Ability(9)
    speed = 20
    swim_speed = 0
    fly_speed = 120
    climb_speed = 0
    hp_max = 248
    hit_dice = "16d20"
    spells = []


class Roper(Monster):
    """
    False Appearance.
      While the roper remains motionless, it is indistinguishable from a
      normal cave formation, such as a stalagmite.
    Grasping Tendrils.
      The roper can have up to six tendrils at a time. Each tendril can be
      attacked (AC 20; 10 hit points; immunity to poison and psychic
      damage). Destroying a tendril deals no damage to the roper, which can
      extrude a replacement tendril on its next turn. A tendril can also be
      broken if a creature takes an action and succeeds on a DC 15 Strength
      check against it.
    Spider Climb.
      The roper can climb difficult surfaces, including upside down on
      ceilings, without needing to make an ability check.
    Multiattack.
      The roper makes four attacks with its tendrils, uses Reel, and makes
      one attack with its bite.
    Bite.
      Melee Weapon Attack: +7 to hit, reach 5 ft., one target. Hit: 22 (4d8
      + 4) piercing damage.
    Tendril.
      Melee Weapon Attack: +7 to hit, reach 50 ft., one creature. Hit: The
      target is grappled (escape DC 15). Until the grapple ends, the target
      is restrained and has disadvantage on Strength checks and Strength
      saving throws, and the roper can't use the same tendril on another
      target.
    Reel.
      The roper pulls each creature grappled by it up to 25 ft. straight
      toward it.
    """
    name = "Roper"
    description = "Large monstrosity, neutral evil"
    challenge_rating = 5
    armor_class = 20
    skills = "Perception +6, Stealth +5"
    senses = "Darkvision 60 ft., Passive Perception 16"
    languages = ""
    strength = Ability(18)
    dexterity = Ability(8)
    constitution = Ability(17)
    intelligence = Ability(7)
    wisdom = Ability(16)
    charisma = Ability(6)
    speed = 10
    swim_speed = 0
    fly_speed = 0
    climb_speed = 10
    hp_max = 93
    hit_dice = "11d10"
    spells = []


class RugofSmothering(Monster):
    """
    Antimagic Susceptibility.
      The rug is incapacitated while in the area of an antimagic field. If
      targeted by dispel magic, the rug must succeed on a Constitution
      saving throw against the caster's spell save DC or fall unconscious
      for 1 minute.
    Damage Transfer.
      While it is grappling a creature, the rug takes only half the damage
      dealt to it, and the creature grappled by the rug takes the other
      half.
    False Appearance.
      While the rug remains motionless, it is indistinguishable from a
      normal rug.
    Smother.
      Melee Weapon Attack: +5 to hit, reach 5 ft., one Medium or smaller
      creature. Hit: The creature is grappled (escape DC 13). Until this
      grapple ends, the target is restrained, blinded, and at risk of
      suffocating, and the rug can't smother another target. In addition, at
      the start of each of the target's turns, the target takes 10 (2d6 + 3)
      bludgeoning damage.
    """
    name = "Rug of Smothering"
    description = "Large construct, unaligned"
    challenge_rating = 2
    armor_class = 12
    skills = ""
    senses = "Blindsight 60 ft. (blind beyond this radius), Passive Perception 6"
    languages = ""
    strength = Ability(17)
    dexterity = Ability(14)
    constitution = Ability(10)
    intelligence = Ability(1)
    wisdom = Ability(3)
    charisma = Ability(1)
    speed = 10
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 33
    hit_dice = "6d10"
    spells = []


class RustMonster(Monster):
    """
    Iron Scent.
      The rust monster can pinpoint, by scent, the location of ferrous metal
      within 30 feet of it.
    Rust Metal.
      Any nonmagical weapon made of metal that hits the rust monster
      corrodes. After dealing damage, the weapon takes a permanent and
      cumulative -1 penalty to damage rolls. If its penalty drops to -5, the
      weapon is destroyed. Non magical ammunition made of metal that hits
      the rust monster is destroyed after dealing damage.
    Bite.
      Melee Weapon Attack: +3 to hit, reach 5 ft., one target. Hit: 5 (1d8 +
      1) piercing damage.
    Antennae.
      The rust monster corrodes a nonmagical ferrous metal object it can see
      within 5 feet of it. If the object isn't being worn or carried, the
      touch destroys a 1-foot cube of it. If the object is being worn or
      carried by a creature, the creature can make a DC 11 Dexterity saving
      throw to avoid the rust monster's touch.
      
      If the object touched is either metal armor or a metal shield being
      worn or carried, its takes a permanent and cumulative -1 penalty to
      the AC it offers. Armor reduced to an AC of 10 or a shield that drops
      to a bonus is destroyed. If the object touched is a held metal weapon,
      it rusts as described in the Rust Metal trait.
    """
    name = "Rust Monster"
    description = "Medium monstrosity, unaligned"
    challenge_rating = 0.5
    armor_class = 14
    skills = ""
    senses = "Darkvision 60 ft., Passive Perception 11"
    languages = ""
    strength = Ability(13)
    dexterity = Ability(12)
    constitution = Ability(13)
    intelligence = Ability(2)
    wisdom = Ability(13)
    charisma = Ability(6)
    speed = 40
    swim_speed = 0
    fly_speed = 0
    climb_speed = 0
    hp_max = 27
    hit_dice = "5d8"
    spells = []



from dungeonsheets.features.features import Feature


# PHB
class Actor(Feature):
    """Skilled at mimicry and dramatics, you gain the following benefits:

    - Increase your Charisma score by 1, to a maximum of 20.
    - You have an advantage on Charisma (Deception) and Charisma
      (Performance) checks when trying to pass yourself off as a
      different person.
    - You can mimic the speech of another person or the sounds made by
      other creatures. You must have heard the person speaking, or
      heard the creature make the sound, for at least 1 minute. A
      successful Wisdom (Insight) check contested by your Charisma
      (Deception) check allows a listener to determine that the effect
      is faked.

    """

    name = "Actor"
    source = "Feats"


class Alert(Feature):
    """Always on the lookout for danger, you gain the following benefits:

    - You gain a +5 bonus to initiative.
    - You can’t be surprised while you are conscious.
    - Other creatures don’t gain advantage on attack rolls against you
      as a result of being hidden from you.

    """

    name = "Alert"
    source = "Feats"


class Athlete(Feature):
    """You have undergone extensive physical training to gain the following
    benefits:

    - Increase your Strength or Dexterity score by 1, to a maximum of
      20.
    - When you are prone, standing up uses only 5 feet of your
      movement.
    - Climbing doesn’t halve your speed.
    - You can make a running long jump or a running high jump after
      moving only 5 feet on foot, rather than 10 feet.

    """

    name = "Athlete"
    source = "Feats"


class Charger(Feature):
    """When you use your action to Dash, you can use a bonus action to
    make one melee weapon attack or to shove a creature.

    If you move at least 10 feet in a straight line immediately before
    taking this bonus action, you either gain a +5 bonus to the
    attack’s damage roll (if you chose to make a melee attack and hit)
    or push the target up to 10 feet away from you (if you chose to
    shove and you succeed).

    """

    name = "Charger"
    source = "Feats"


class CrossbowExpert(Feature):
    """Thanks to extensive practice with the crossbow, you gain the following
    benefits:

    - You ignore the loading quality of crossbows with which you are
      proficient.
    - Being within 5 feet of a hostile creature doesn’t impose
      disadvantage on your ranged attack rolls.
    - When you use the Attack action and attack with a one-handed
      weapon, you can use a bonus action to attack with a loaded hand
      crossbow you are holding.

    """

    name = "Crossbow Expert"
    source = "Feats"


class DefensiveDuelist(Feature):
    """When you are wielding a finesse weapon with which you are
    proficient and another creature hits you with a melee attack, you
    can use your reaction to add your proficiency bonus to your AC for
    that attack, potentially causing the attack to miss you.

    **Prerequisite:** Dexterity 13 or higher

    """

    name = "Defensive Duelist"
    source = "Feats"


class DualWielder(Feature):
    """You master fighting with two weapons, gaining the following benefits:

    - You gain a +1 bonus to AC while you are wielding a separate
      melee weapon in each hand.
    - You can use two-weapon fighting even when the one-handed melee
      weapons you are wielding aren’t light.
    - You can draw or stow two one-handed weapons when you would
      normally be able to draw or stow only one.

    """

    name = "Dual Wielder"
    source = "Feats"


class DungeonDelver(Feature):
    """Alert to the hidden traps and secret doors found in many dungeons, you
    gain the following benefits:

    - When you roll a Hit Die to regain hit points, the minimum number
      of hit points you regain from the roll equals twice your
      Constitution modifier (minimum of 2).
    - You have advantage on Wisdom (Perception) and Intelligence
      (Investigation) checks made to detect the presence of secret
      doors.
    - You have advantage on saving throws made to avoid or resist
      traps.
    - You have resistance to the damage dealt by traps.
    - You can search for traps while travelling at a normal pace,
      instead of only at a slow pace.

    """

    name = "Dungeon Delver"
    source = "Feats"


class Durable(Feature):
    """Hardy and resilient, you gain the following benefits:

    - Increase your Constitution score by 1, to a maximum of 20.
    - When you roll a Hit Die to regain hit points, the minimum number
      of hit points you regain from the roll equals twice your
      Constitution modifier (minimum of 2).

    """

    name = "Durable"
    source = "Feats"


class ElementalAdept(Feature):
    """When you gain this feat, choose one of the following damage types:
    acid, cold, fire, lightning, or thunder.

    Spells you cast ignore resistance to damage of the chosen type. In
    addition, when you roll damage for a spell you cast that deals
    damage of that type, you can treat any 1 on a damage die as a 2.

    You can select this feat multiple times. Each time you do so, you
    must choose a different damage type.

    **Prerequisite:** The ability to cast at least one spell

    """

    name = "Elemental Adept"
    source = "Feats"


class Grappler(Feature):
    """You’ve developed the skills necessary to hold your own in close-quarters
    grappling. You gain the following benefits:

    - You have advantage on attack rolls against a creature you are
      grappling.
    - You can use your action to try to pin a creature grappled by
      you. To do so, make another grapple check. If you succeed, you
      and the creature are both restrained until the grapple ends.
    - Creatures that are one size larger than you don’t automatically
      succeed on checks to escape your grapple.

    **Prerequisite:** Strength 13 or higher

    """

    name = "Grappler"
    source = "Feats"


class GreatWeaponMaster(Feature):
    """You've learned to put the weight of a weapon to your advantage, letting
    its momentum empower your strikes. You gain the following benefits:

    - On your turn, when you score a critical hit with a melee weapon
      or reduce a creature to 0 hit points with one, you can make one
      melee weapon attack as a bonus action.
    - Before you make a melee attack with a heavy weapon that you are
      proficient with, you can choose to take a -5 penalty to the
      attack roll. If the attack hits, you add +10 to the attack's
      damage

    """

    name = "Great Weapon Master"
    source = "Feats"


class Healer(Feature):
    """You are an able physician, allowing you to mend wounds quickly and get
    your allies back in the fight. You gain the following benefits:

    - When you use a healer’s kit to stabilize a dying creature, that
      creature also regains 1 hit point.
    - As an action, you can spend one use of a healer’s kit to tend to
      a creature and restore 1d6 + 4 hit points to it, plus additional
      hit points equal to the creature’s maximum number of Hit
      Dice. The creature can’t regain hit points from this feat again
      until it finishes a short or long rest.

    """

    name = "Healer"
    source = "Feats"


class HeavilyArmored(Feature):
    """You have trained to master the use of heavy armor, gaining
    the following benefits:

    - Increase your Strength score by 1, to a maximum of 20.
    - You gain proficiency with heavy armor.

    **Prerequisite:** Proficiency with medium armor

    """

    name = "Heavily Armored"
    source = "Feats"


class HeavyArmorMaster(Feature):
    """You can use your armor to deflect strikes that would kill others. You
    gain the following benefits:

    - Increase your Strength score by 1, to a maximum of 20.
    - While you are wearing heavy armor, bludgeoning, piercing, and
      slashing damage that you take from non magical weapons is
      reduced by 3.

    **Prerequisite:** Proficiency with heavy armor

    """

    name = "Heavy Armor Master"
    source = "Feats"


class InspiringLeader(Feature):
    """You can spend 10 minutes inspiring your companions, shoring up
    their resolve to fight. When you do so, choose up to six friendly
    creatures (which can include yourself) within 30 feet of you who
    can see or hear you and who can understand you. Each creature can
    gain temporary hit points equal to your level + your Charisma
    modifier.

    A creature can’t gain temporary hit points from this feat again
    until it has finished a short or long rest.

    **Prerequisite:** Charisma 13 or higher

    """

    name = "Inspiring Leader"
    source = "Feats"


class KeenMind(Feature):
    """You have a mind that can track time, direction, and detail with uncanny
    precision. You gain the following benefits:

    - Increase your Intelligence score by 1, to a maximum of 20.
    - You always know which way is north.
    - You always know the number of hours left before the next sunrise
      or sunset.
    - You can accurately recall anything you have seen or heard within
      the past month.

    """

    name = "Keen Mind"
    source = "Feats"


class LightlyArmored(Feature):
    """You have trained to master the use of light armor, gaining the following
    benefits:

    - Increase your Strength or Dexterity score by 1, to a maximum of
      20.
    - You gain proficiency with light armor.

    """

    name = "Lightly Armored"
    source = "Feats"


class Linguist(Feature):
    """You have studied languages and codes, gaining the following benefits:

    - Increase your Intelligence score by 1, to a maximum of 20.
    - You learn three languages of your choice.
    - You can ably create written ciphers. Others can’t decipher a
      code you create unless you teach them, they succeed on an
      Intelligence check (DC equal to your Intelligence score + your
      proficiency bonus), or they use magic to decipher it.

    """

    name = "Linguist"
    source = "Feats"


class Lucky(Feature):
    """You have inexplicable luck that seems to kick in at just the right
    moment.

    You have 3 luck points. Whenever you make an attack roll, an ability check,
    or a saving throw, you can spend one luck point to roll an additional d20.
    You can choose to spend one of your luck points after you roll the die, but
    before the outcome is determined. You choose which of the d20s is used for
    the attack roll, ability check, or saving throw.

    You can also spend one luck point when an attack roll is made against you.
    Roll a d20, and then choose whether the attack uses the attacker’s roll or
    yours.

    If more than one creature spends a luck point to influence the outcome of
    a roll, the points cancel each other out; no additional dice are rolled.

    You regain your expended luck points when you finish a long rest.
    """

    name = "Lucky"
    source = "Feats"


class MageSlayer(Feature):
    """You have practiced techniques useful in melee combat against
    spellcasters, gaining the following benefits:

    - When a creature within 5 feet of you casts a spell, you can use
      your reaction to make a melee weapon attack against that
      creature.
    - When you damage a creature that is concentrating on a spell,
      that creature has disadvantage on the saving throw it makes to
      maintain its concentration.
    - You have advantage on saving throws against spells cast by
      creatures within 5 feet of you.

    """

    name = "Mage Slayer"
    source = "Feats"


class MagicInitiate(Feature):
    """Choose a class: bard, cleric, druid, sorcerer, warlock, or wizard.

    You learn two cantrips of your choice from that class’s spell list.

    In addition, choose one 1st-level spell from that same list. You learn that
    spell and can cast it at its lowest level. Once you cast it, you must finish
    a long rest before you can cast it again.

    Your spellcasting ability for these spells depends on the class you chose:
    Charisma for bard, sorcerer, or warlock; Wisdom for cleric or druid: or
    Intelligence for wizard.
    """

    name = "Magic Initiate"
    source = "Feats"


class MartialAdept(Feature):
    """You have martial training that allows you to perform special combat
    maneuvers. You gain the following benefits:

    - You learn two maneuvers of your choice from among those
      available to the Battle Master archetype in the fighter
      class. If a maneuver you use requires your target to make a
      saving throw to resist the maneuver’s effects, the saving throw
      DC equals 8 + your proficiency bonus + your Strength or
      Dexterity modifier (your choice).
    - If you already have superiority dice, you gain one more;
      otherwise, you have one superiority die, which is a d6. This die
      is used to fuel your maneuvers. A superiority die is expended
      when you use it. You regain your expended superiority dice when
      you finish a short or long rest.

    """

    name = "Martial Adept"
    source = "Feats"


class MediumArmorMaster(Feature):
    """You have practiced moving in medium armor to gain the following benefits:

    - Wearing medium armor doesn’t impose disadvantage on your
      Dexterity (Stealth) checks.
    - When you wear medium armor, you can add 3, rather than 2, to
      your AC if you have a Dexterity of 16 or higher.

    **Prerequisite:** Prerequisite: Proficiency with medium armor

    """

    name = "Medium Armor Master"
    source = "Feats"


class Mobile(Feature):
    """You are exceptionally speedy and agile. You gain the following benefits:

    - Your speed increases by 10 feet.
    - When you use the Dash action, difficult terrain doesn’t cost you
      extra movement on that turn.
    - When you make a melee attack against a creature, you don’t
      provoke opportunity attacks from that creature for the rest o f
      the turn, whether you hit or not.

    """

    name = "Mobile"
    source = "Feats"


class ModeratelyArmored(Feature):
    """You have trained to master the use of medium armor and shields, gaining
    the following benefits:

    - Increase your Strength or Dexterity score by 1, to a maximum of
      20.
    - You gain proficiency with medium armor and shields.

    **Prerequisite:** Proficiency with light armor

    """

    name = "Moderately Armored"
    source = "Feats"


class MountedCombatant(Feature):
    """You are a dangerous foe to face while mounted. While you are mounted and
    aren’t incapacitated, you gain the following benefits:

    - You have advantage on melee attack rolls against any unmounted
      creature that is smaller than your mount.
    - You can force an attack targeted at your mount to target you instead.
    - If your mount is subjected to an effect that allows it to make a
      Dexterity saving throw to take only half damage, it instead
      takes no damage if it succeeds on the saving throw, and only
      half damage if it fails.

    """

    name = "Mounted Combatant"
    source = "Feats"


class Observant(Feature):
    """Quick to notice details of your environment, you gain the following
    benefits:

    - Increase your Intelligence or Wisdom score by 1, to a maximum of
      20.
    - If you can see a creature’s mouth while it is speaking a
      language you understand, you can interpret what it’s saying by
      reading its lips.
    - You have a +5 bonus to your passive Wisdom (Perception) and
      passive Intelligence (Investigation) scores.

    """

    name = "Observant"
    source = "Feats"


class PolearmMaster(Feature):
    """You can keep your enemies at bay with reach weapons. You gain
    the following benefits:

    - When you take the Attack action and attack with only a glaive,
      halberd, or quarterstaff, you can use a bonus action to make a
      melee attack with the opposite end of the weapon. The weapon’s
      damage die for this attack is a d4, and the attack deals
      bludgeoning damage.
    - While you are wielding a glaive, halberd, pike, or quarterstaff,
      other creatures provoke an opportunity attack from you when they
      enter your reach.

    """

    name = "Polearm Master"
    source = "Feats"


class Resilient(Feature):
    """Choose one ability score. You gain the following benefits:

    - Increase the chosen ability score by 1, to a maximum of 20.
    - You gain proficiency in saving throws using the chosen ability.

    """

    name = "Resilient"
    source = "Feats"
    needs_implementation = True


class RitualCaster(Feature):
    """You have learned a number of spells that you can cast as
    rituals. These spells are written in a ritual book, which you must
    have in hand while casting one of them.

    When you choose this feat, you acquire a ritual book holding two
    1st-level spells of your choice. Choose one of the following
    classes: bard, cleric, druid, sorcerer, warlock, or wizard. You
    must choose your spells from that class’s spell list, and the
    spells you choose must have the ritual tag.  The class you choose
    also determines your spellcasting ability for these spells:
    Charisma for bard, sorcerer, or warlock; Wisdom for cleric or
    druid; or Intelligence for wizard.

    If you come across a spell in written form, such as a magical
    spell scroll or a wizard’s spellbook, you might be able to add it
    to your ritual book.  The spell must be on the spell list for the
    class you chose, the spell’s level can be no higher than half your
    level (rounded up), and it must have the ritual tag. The process
    of copying the spell into your ritual book takes 2 hours per level
    of the spell, and costs 50 gp per level. The cost represents
    material components you expend as you experiment with the spell to
    master it, as well as the fine inks you need to record it.

    Prerequisite: Intelligence or Wisdom 13 or higher

    """

    name = "Ritual Caster"
    source = "Feats"


class SavageAttacker(Feature):
    """Once per turn when you roll damage for a melee weapon attack, you
    can reroll the weapon’s damage dice and use either total.

    """

    name = "Savage Attacker"
    source = "Feats"


class Sentinel(Feature):
    """You have mastered techniques to take advantage of every drop in any
    enemy’s guard, gaining the following benefits:

    - When you hit a creature with an opportunity attack, the
      creature’s speed becomes 0 for the rest of the turn.
    - Creatures within 5 feet of you provoke opportunity attacks from
      you even if they take the Disengage action before leaving your
      reach.
    - When a creature within 5 feet of you makes an attack against a
      target other than you (and that target doesn’t have this feat),
      you can use your reaction to make a melee weapon attack against
      the attacking creature.

    """

    name = "Sentinel"
    source = "Feats"


class Sharpshooter(Feature):
    """You have mastered ranged weapons and can make shots that others find
    impossible. You gain the following benefits:

    - Attacking at long range doesn't impose disadvantage on your ranged weapon
      attack rolls.
    - Your ranged weapon attacks ignore half cover and three-quarters
      cover.
    - Before you make an attack with a ranged weapon that you are
      proficient with, you can choose to take a - 5 penalty to the
      attack roll. If the attack hits, you add +10 to the attack’s
      damage.

    """

    name = "Sharpshooter"
    source = "Feats"


class ShieldMaster(Feature):
    """You use shields not just for protection but also for offense. You gain
    the following benefits while you are wielding a shield:

    - If you take the Attack action on your turn, you can use a bonus
      action to try to shove a creature within 5 feet of you with your
      shield.
    - If you aren’t incapacitated, you can add your shield’s AC bonus
      to any Dexterity saving throw you make against a spell or other
      harmful effect that targets only you.
    - If you are subjected to an effect that allows you to make a
      Dexterity saving throw to take only half damage, you can use
      your reaction to take no damage if you succeed on the saving
      throw, interposing your shield between yourself and the source
      of the effect.

    """

    name = "Shield Master"
    source = "Feats"


class Skilled(Feature):
    """You gain proficiency in any combination of three skills or tools of
    your choice.

    """

    name = "Skilled"
    source = "Feats"


class Skulker(Feature):
    """You are expert at slinking through shadows. You gain the following
    benefits:

    - You can try to hide when you are lightly obscured from the
      creature from which you are hiding.
    - When you are hidden from a creature and miss it with a ranged
      weapon attack, making the attack doesn't reveal your position.
    - Dim light doesn’t impose disadvantage on your Wisdom
      (Perception) checks relying on sight.

    Prerequisite: Dexterity 13 or higher

    """

    name = "Skulker"
    source = "Feats"


class SpellSniper(Feature):
    """You have learned techniques to enhance your attacks with certain kinds of
    spells, gaining the following benefits:

    - When you cast a spell that requires you to make an attack roll,
      the spell’s range is doubled.
    - Your ranged spell attacks ignore half cover and three-quarters
      cover.
    - You learn one cantrip that requires an attack roll.

    Choose the cantrip from the bard, cleric, druid, sorcerer,
    warlock, or wizard spell list. Your spellcasting ability for this
    cantrip depends on the spell list you chose from: Charisma for
    bard, sorcerer, or warlock; Wisdom for cleric or druid; or
    Intelligence for wizard.

    Prerequisite: The ability to cast at least one spell

    """

    name = "Spell Sniper"
    source = "Feats"


class TavernBrawler(Feature):
    """Accustomed to rough-and-tumble fighting using whatever weapons happen to
    be at hand, you gain the following benefits:

    - Increase your Strength or Constitution score by 1, to a maximum
      of 20.
    - You are proficient with improvised weapons and unarmed strikes.
    - Your unarmed strike uses a d4 for damage.
    - When you hit a creature with an unarmed strike or an improvised
      weapon on your turn, you can use a bonus action to attempt to
      grapple the target.

    """

    name = "Tavern Brawler"
    source = "Feats"


class Tough(Feature):
    """Your hit point maximum increases by an amount equal to twice your
    level when you gain this feat. Whenever you gain a level
    thereafter, your hit point maximum increases by an additional 2
    hit points.

    """

    name = "Tough"
    source = "Feats"


class WarCaster(Feature):
    """You have practiced casting spells in the midst of combat, learning
    techniques that grant you the following benefits:

    - You have advantage on Constitution saving throws that you make
      to maintain your concentration on a spell when you take damage.
    - You can perform the somatic components of spells even when you
      have weapons or a shield in one or both hands.
    - When a hostile creature’s movement provokes an opportunity
      attack from you, you can use your reaction to cast a spell at
      the creature, rather than making an opportunity attack. The
      spell must have a casting time of 1 action and must target only
      that creature.

    **Prerequisite:** The ability to cast at least one spell

    """

    name = "War Caster"
    source = "Feats"


class WeaponMaster(Feature):
    """You have practiced extensively with a variety of weapons, gaining
    the following benefits:

    - Increase your Strength or Dexterity score by 1, to a maximum of
      20.
    - You gain proficiency with four weapons of your choice.

    """

    name = "Weapon Master"
    source = "Feats"


# UA
class BarbedHide(Feature):
    """One of your ancestors was a barbed devil or other spiky
    fiend. Barbs protrude from your head. You gain the following
    benefits:

    - Increase your Constitution or Charisma score by 1, up to a
      maximum of 20.
    - As a bonus action, you can cause small barbs to protrude all
      over your body or cause them to retract. At the start of each of
      your turns while the barbs are out, you deal 1d6 piercing damage
      to any creature grappling you or any creature grappled by you.
    - You gain proficiency in the Intimidation skill. If you’re
      already proficient in it, your proficiency bonus is doubled for
      any check you make with it.

    **Prerequisite:** Tiefling

    """

    name = "Barbed Hide"
    source = "Feats"


class BountifulLuckUA(Feature):
    """Whenever an ally you can see within 30 feet of you rolls a 1 on the d20
    for an attack roll, an ability check, or a saving throw, you can use your
    reaction to let the ally reroll the die. The ally must use the new roll.

    **Prerequisite:** Halfling
    """

    name = "Bountiful Luck (UA)"
    source = "Feats"


class CritterFriend(Feature):
    """Your friendship with animals mystically deepens. You gain the
    following benefits:

    - You gain proficiency in the Animal Handling skill. If you’re
      already proficient in it, your proficiency bonus is doubled for
      any check you make with it.
    - You learn the speak with animals spell and can cast it at will,
      without expending a spell slot. You also learn the animal
      friendship spell, and you can cast it once with this feat,
      without expending a spell slot. You regain the ability to cast
      it in this way when you finish a long rest.  Intelligence is
      your spellcasting ability for these spells.

    **Prerequisite:** Gnome (Forest)

    """

    name = "Critter Friend"
    source = "Feats"


class DragonWings(Feature):
    """You sprout draconic wings. With your wings, you have a flying speed of
    20 feet if you aren’t wearing heavy armor and aren’t exceeding your carrying
    capacity.

    **Prerequisite:** Dragonborn
    """

    name = "Dragon Wings"
    source = "Feats"


class EverybodysFriend(Feature):
    """You develop your magnetic personality to ease your way through the world.
    You gain the following benefits:

    - Increase your Charisma score by 1, up to a maximum of 20.
    - You gain proficiency in the Deception and Persuasion skills. If
      you're already proficient in either skill, your proficiency
      bonus is doubled for any check you make with that skill.

    **Prerequisite:** Half-elf

    """

    name = "Everybody's Friend"
    source = "Feats"


class GrudgeBearer(Feature):
    """You have a deep hatred for a particular kind of creature. Choose
    your foes, a type of creature to bear the burden of your wrath:
    aberrations, beasts, celestials, constructs, dragons, elementals,
    fey, fiends, giants, monstrosities, oozes, plants, or
    undead. Alternatively, you can choose two races of humanoid (such
    as gnolls and orcs). You gain the following benefits:

    - Increase your Strength, Constitution, or Wisdom score by 1, to a
      maximum of 20.
    - During the first round of any combat against your chosen foes,
      your attack rolls against any of them have advantage.
    - When any of your chosen foes makes an opportunity attack against
      you, it makes the attack roll with disadvantage.
    - Whenever you make an Intelligence (Arcana, History, Nature, or
      Religion) check to recall information about your chosen foes,
      you add double your proficiency bonus to the check, even if
      you’re not normally proficient.

    **Prerequisite:** Dwarf

    """

    name = "Grudge-Bearer"
    source = "Feats"


class HumanDetermination(Feature):
    """You are filled with a determination that can draw the unreachable
    within your reach. You gain the following benefits:

    - Increase one ability score of your choice by 1, to a maximum of
      20.

    - When you make an attack roll, an ability check, or a saving
      throw, you can do so with advantage. Once you use this ability,
      you can’t use it again until you finish a short or long rest.

    **Prerequisite:** Human

    """

    name = "Human Determination"
    source = "Feats"


class OrcishAggression(Feature):
    """As a bonus action, you can move up to your speed toward an enemy of
    your choice that you can see or hear. You must end this move
    closer to the enemy than you started.

    **Prerequisite:** Half-orc

    """

    name = "Orcish Aggression"
    source = "Feats"


class WonderMaker(Feature):
    """You master the tinker techniques of your people. You gain
    the following benefits:

    - Increase your Dexterity or Intelligence score by 1, to a maximum
      of 20.
    - When you make a check using your proficiency with tinker’s
      tools, you add double your proficiency bonus to the check.
    - When you make a device with your Tinker trait, you have the
      following additional options for what you make:

    **Alarm.** This device senses when a creature moves to within 15 feet of it
    without speaking aloud a password chosen when you create it. One round after
    a creature moves into range, the alarm makes a shrill ringing that lasts for
    1 minute and can be heard from up to 300 feet away.

    **Calculator.** This device makes doing sums easy.

    **Lifter.** This device can be used as a block and tackle, allowing its user
    to hoist five times the weight the user can normally lift.

    **Timekeeper.** This pocket watch keeps accurate time.

    **Weather Sensor.** When used as an action, this device predicts weather
    conditions in a 1-mile radius over the next 4 hours, showing one symbol
    (clouds, sun/moon, rain, or snow) for each hour.

    **Prerequisite:** Gnome (rock)

    """

    name = "Wonder Maker"
    source = "Feats"


# XGE
class BountifulLuck(Feature):
    """Your people have extraordinary luck, which you have learned to
    mystically lend to your companions when you see them
    falter. You're not sure how you do it; you just wish it, and it
    happens. Surely a sign of fortune’s favor! When an ally you can
    see within 30 feet of you rolls a 1 on the d20 for an attack roll,
    an ability check, or a saving throw, you can use your reaction to
    let the ally reroll the die. The ally must use the new roll. When
    you use this ability, you can’t use your Lucky racial trait before
    the end of your next turn.

    **Prerequisite:** Halfling

    """

    name = "Bountiful Luck"
    source = "Feats"


class DragonFear(Feature):
    """When angered, you radiate menace. You gain the following benefits:

    - Increase your Strength, Constitution or Charisma score by 1, up
      to a maximum of 20.
    - Instead of exhaling destructive energy, you can expend a use of
      your Breath Weapon trait to roar, forcing each creature of your
      choice within 30 feet of you to make a Wisdom saving throw (DC 8
      + your proficiency bonus + your Charisma modifier). A target
      automatically succeeds if it can’t hear or see you. On a failed
      save, a target becomes frightened of you for 1 minute. If the
      frightened target takes any damage, it can repeat the saving
      throw, ending the effect on itself on a success.

    **Prerequisite:** Dragonborn

    """

    name = "Dragon Fear"
    source = "Feats"


class DragonHide(Feature):
    """You manifest scales and claws reminiscent of your draconic
    ancestors. You gain the following benefits:

    - Increase your Strength, Constitution or Charisma score by 1, up
      to a maximum of 20.
    - Your scales harden. While you aren’t wearing armor, you can
      calculate your AC as 13 + your Dexterity modifier. You can use a
      shield and still gain this benefit.
    - You grow retractable claws from the tips of your
      fingers. Extending or retracting the claws requires no
      action. The claws are natural weapons, which you can use to make
      unarmed strikes. If you hit with them, you deal slashing damage
      equal to 1d4 + your Strength modifier, instead of the normal
      bludgeoning damage for an unarmed strike.

    **Prerequisite:** Dragonborn

    """

    name = "Dragon Hide"
    source = "Feats"


class DrowHideMagic(Feature):
    """You learn more of the magic typical of dark elves. You learn the
    detect magic spell and can cast it at will, without expending a
    spell slot. You also learn levitate and dispel magic, each of
    which you can cast once without expending a spell slot. You regain
    the ability to cast those two spells in this way when you finish a
    long rest.

    Charisma is your spellcasting ability for all three spells.

    **Prerequisite:** Elf (drow)

    """

    name = "Drow Hide Magic"
    source = "Feats"


class DwarvenFortitude(Feature):
    """You have the blood of dwarf heroes flowing through your veins. You gain
    the following benefits:

    - Increase your Constitution score by 1, up to a maximum of 20.
    - Whenever you take the Dodge action in combat, you can spend one
      Hit Die to heal yourself. Roll the die, add your Constitution
      modifier, and regain a number of hit points equal to the total
      (minimum of 1).

    **Prerequisite:** Dwarf

    """

    name = "Dwarven Fortitude"
    source = "Feats"


class ElvenAccuracy(Feature):
    """The accuracy of elves is legendary, especially that of elf archers
    and spellcasters. You have uncanny aim with attacks that rely on
    precision rather than brute force. You gain the following
    benefits:

    - Increase your Dexterity, Intelligence, Wisdom, or Charisma score
      by 1, to a maximum of 20.
    - Whenever you have advantage on an attack roll using Dexterity,
      Intelligence, Wisdom, or Charisma, you can reroll one of the
      dice once.

    **Prerequisite:** Elf or half-elf

    """

    name = "Elven Accuracy"
    source = "Feats"


class FadeAway(Feature):
    """Your people are clever, with a knack for illusion magic. You have
    learned a magical trick for fading away when you suffer harm. You
    gain the following benefits:

    - Increase your Dexterity or Intelligence score by 1, to a maximum
      of 20.
    - Immediately after you take damage, you can use a reaction to
      magically become invisible until the end of your next turn or
      until you attack, deal damage, or force someone to make a saving
      throw. Once you use this ability, you can’t do so again until
      you finish a short or long rest.

    **Prerequisite:** Gnome

    """

    name = "Fade Away"
    source = "Feats"


class FeyTeleportation(Feature):
    """Your study of high elven lore has unlocked fey power that few other elves
    possess, except your eladrin cousins. Drawing on your fey ancestry, you can
    momentarily stride through the Feywild to shorten your path from one place
    to another. You gain the following benefits:

    - Increase your Intelligence or Charisma score by 1, to a maximum
      of 20.
    - You learn to speak, read, and write Sylvan.
    - You learn the misty step spell and can cast it once without
      expending a spell slot. You regain the ability to cast it in
      this way when you finish a short or long rest. Intelligence is
      your spellcasting ability for this spell.

    **Prerequisite:** Elf (high)

    """

    name = "Fey Teleportation"
    source = "Feats"


class FlamesOfPhlegethos(Feature):
    """You learn to call on hellfire to serve your commands. You gain
    the following benefits:

    - Increase your Intelligence or Charisma score by 1, to a maximum
      of 20.
    - When you roll fire damage for a spell you cast, you can reroll
      any roll of 1 on the fire damage dice, but you must use the new
      roll, even if it is another 1.
    - Whenever you cast a spell that deals fire damage, you can cause
      flames to wreathe you until the end of your next turn. The
      flames don’t harm you or your possessions, and they shed bright
      light out to 30 feet and dim light for an additional 30
      feet. While the flames are present, any creature within 5 feet
      of you that hits you with a melee attack takes 1d4 fire damage.

    **Prerequisite:** Tiefling

    """

    name = "Flames of Phlegethos"
    source = "Feats"


class InfernalConstitution(Feature):
    """Fiendish blood runs strong in you, unlocking a resilience akin to
    that possessed by some fiends. You gain the following benefits:

    - Increase your Constitution score by 1, to a maximum of 20.
    - You have resistance to cold damage and poison damage.
    - You have advantage on saving throws against being poisoned.

    **Prerequisite:** Tiefling

    """

    name = "Infernal Constitution"
    source = "Feats"


class OrchishFury(Feature):
    """Your inner fury burns tirelessly. You gain the following benefits:

    - Increase your Strength or Constitution score by 1, to a maximum
      of 20.
    - When you hit with an attack using a simple or martial weapon,
      you can roll one of the weapon’s damage dice an additional time
      and add it as extra damage of the weapon’s damage type. Once you
      use this ability, you can’t use it again until you finish a
      short or long rest.
    - Immediately after you use your Relentless Endurance trait, you
      can use your reaction to make one weapon attack.

    **Prerequisite:** Half-orc

    """

    name = "Orchish Fury"
    source = "Feats"


class Prodigy(Feature):
    """You have a knack for learning new things. You gain the following
    benefits:

    - You gain one skill proficiency of your choice, one tool
      proficiency of your choice, and fluency in one language of your
      choice.
    - Choose one skill in which you have proficiency. You gain
      expertise with that skill, which means your proficiency bonus is
      doubled for any ability check you make with it. The skill you
      choose must be one that isn’t already benefiting from a feature,
      such as Expertise, that doubles your proficiency bonus.

    **Prerequisite:** Half-elf, half-orc, or human

    """

    name = "Prodigy"
    source = "Feats"


class SecondChance(Feature):
    """Fortune favors you when someone tries to strike you. You gain
    the following benefits:

    - Increase your Dexterity, Constitution or Charisma score by 1, to
      a maximum of 20.
    - When a creature you can see hits you with an attack roll, you
      can use your reaction to force that creature to reroll. Once you
      use this ability, you can’t use it again until you roll
      initiative at the start of combat or until you finish a short or
      long rest.

    **Prerequisite:** Halfling

    """

    name = "Second Chance"
    source = "Feats"


class SquatNimbleness(Feature):
    """You are uncommonly nimble for your race. You gain the following benefits:

    - Increase your Strength or Dexterity score by 1, to a maximum of
      20.
    - Increase your walking speed by 5 feet.
    - You gain proficiency in the Acrobatics or Athletics skill (your
      choice).
    - You have advantage on any Strength (Athletics) or Dexterity
      (Acrobatics) check you make to escape from being grappled.

    **Prerequisite:** Dwarf or a Small race

    """

    name = "Squat Nimbleness"
    source = "Feats"


class WoodElfMagic(Feature):
    """You learn the magic of the primeval woods, which are revered and
    protected by your people. You learn one druid cantrip of your
    choice. You also learn the long strider and pass without trace
    spells, each of which you can cast once without expending a spell
    slot. You regain the ability to cast these two spells in this way
    when you finish a long rest. Wisdom is your spellcasting ability
    for all three spells.

    **Prerequisite:** Elf (wood)

    """

    name = "Wood Elf Magic"
    source = "Feats"



"""Base classes for the various D&D 5e content types."""


import warnings
from abc import ABC
from pathlib import Path

from dungeonsheets.stats import Ability, ArmorClass, Initiative, Speed, Skill
from dungeonsheets.content_registry import find_content


def read(fname):
    return open((Path(__file__).parent / fname).resolve()).read()


__version__ = read("../VERSION").strip()


class Content(ABC):
    """A base class for all D&D 5e content types.

    Every piece of content (e.g. class feature, spell, monster) should
    have this base class in its inheritance tree.

    """
    dungeonsheets_version = __version__
    name = "Generic content"

    @staticmethod
    def _resolve_mechanic(mechanic, SuperClass, warning_message=None):
        """Take a raw entry in a character sheet and turn it into a usable object.

        Eg: spells can be defined in many ways. This function accepts all
        of those options and returns an actual *Spell* class that can be
        used by a character::

            >>> _resolve_mechanic("mage_hand", SuperClass=spells.Spell)

            >>> _resolve_mechanic("mage_hand", SuperClass=None)

            >>> from dungeonsheets import spells
            >>> class MySpell(spells.Spell): pass
            >>> _resolve_mechanic(MySpell, SuperClass=spells.Spell)

            >>> _resolve_mechanic("hocus pocus", SuperClass=spells.Spell)

        The acceptable entries for *mechanic*, in priority order, are:
          1. A subclass of *SuperClass*
          2. A string with the name of defined content
          3. The name of an unknown spell (creates generic object using *factory*)

        *SuperClass* can be ``None`` to match any class, however this will
        raise an exception if more than one content type has this
        name. For example, "shield" can refer to both the armor or the
        spell, so ``_resolve_mechanic("shield")`` will raise an exception.

        Parameters
        ==========
        mechanic : str, type
          The thing to be resolved, either a string with the name of the
          mechanic, or a subclass of *ParentClass* describing the
          mechanic.
        SuperClass : type
          Class to determine whether *mechanic* should just be allowed
          through as is.
        error_message : str, optional
          A string whose ``str.format()`` method (receiving one positional
          argument *mechanic*) will be used for displaying a warning when an
          unknown mechanic is resolved. If omitted, no warning will be
          displayed.

        Returns
        =======
        Mechanic
          A class representing the resolved game mechanic. This will
          likely be a subclass of *SuperClass* if the other parameters are
          well behaved, but this is not enforced.

        """
        is_already_resolved = isinstance(mechanic, type) and issubclass(
            mechanic, SuperClass
        )
        if is_already_resolved:
            Mechanic = mechanic
        elif SuperClass is not None and isinstance(mechanic, SuperClass):
            # Has been instantiated for some reason
            Mechanic = type(Mechanic)
        else:
            try:
                # Retrieve pre-defined mechanic
                valid_classes = [SuperClass] if SuperClass is not None else []
                Mechanic = find_content(mechanic, valid_classes=valid_classes)
            except ValueError:
                # No pre-defined mechanic available
                if warning_message is not None:
                    # Emit the warning
                    msg = warning_message.format(mechanic)
                    warnings.warn(msg)
                else:
                    # Create a generic message so we can make a docstring later.
                    msg = f'Mechanic "{mechanic}" not defined. Please add it.'
                # Create generic mechanic from the factory
                class_name = "".join([s.title() for s in mechanic.split("_")])
                mechanic_name = mechanic.replace("_", " ").title()
                attrs = {"name": mechanic_name, "__doc__": msg, "source": "Unknown"}
                Mechanic = type(class_name, (SuperClass,), attrs)
        return Mechanic
    


class Creature(Content):
    """A thing with stats. Use Monster or Character, not this class
    directly!

    """
    # General attributes
    alignment = "Neutral"
    _race = None
    name = "Generic creature"
    # Hit points
    hp_max = None
    # Base stats (ability scores)
    strength = Ability()
    dexterity = Ability()
    constitution = Ability()
    intelligence = Ability()
    wisdom = Ability()
    charisma = Ability()
    # Numerical things
    armor_class = ArmorClass()
    initiative = Initiative()
    speed = Speed()
    # Proficiencies and Languages
    _saving_throw_proficiencies = tuple()  # use to overwrite class proficiencies
    other_weapon_proficiencies = tuple()  # add to class/race proficiencies
    skill_proficiencies = list()
    skill_expertise = list()
    languages = ""
    senses = ""
    # Skills
    acrobatics = Skill(ability="dexterity")
    animal_handling = Skill(ability="wisdom")
    arcana = Skill(ability="intelligence")
    athletics = Skill(ability="strength")
    deception = Skill(ability="charisma")
    history = Skill(ability="intelligence")
    insight = Skill(ability="wisdom")
    intimidation = Skill(ability="charisma")
    investigation = Skill(ability="intelligence")
    medicine = Skill(ability="wisdom")
    nature = Skill(ability="intelligence")
    perception = Skill(ability="wisdom")
    performance = Skill(ability="charisma")
    persuasion = Skill(ability="charisma")
    religion = Skill(ability="intelligence")
    sleight_of_hand = Skill(ability="dexterity")
    stealth = Skill(ability="dexterity")
    survival = Skill(ability="wisdom")
    # Inventory
    cp = 0
    sp = 0
    ep = 0
    gp = 0
    pp = 0
    equipment = ""
    _weapons = list()
    magic_items = list()
    armor = None
    shield = None
    # Magic
    spellcasting_ability = None
    _spells = list()
    _spells_prepared = list()
    infusions = list()
    # Features IN MAJOR DEVELOPMENT
    custom_features = list()
    feature_choices = list()

    def __init__(self):
        pass

    @property
    def weapons(self):
        return self._weapons.copy()

    @property
    def passive_wisdom(self):
        return self.perception.modifier + 10

    @property
    def abilities(self):
        return [self.strength, self.dexterity, self.constitution,
                self.intelligence, self.wisdom, self.charisma]

    @property
    def skills(self):
        return [self.acrobatics, self.animal_handling, self.arcana,
                self.athletics, self.deception, self.history,
                self.insight, self.intimidation, self.investigation,
                self.medicine, self.nature, self.perception,
                self.performance, self.persuasion, self.religion,
                self.sleight_of_hand, self.stealth, self.survival]



import importlib
import warnings
import json
import re
from functools import lru_cache
import logging
from typing import Union

from pathlib import Path

from dungeonsheets import exceptions

log = logging.getLogger(__file__)


def read_sheet_file(filename: Union[str, Path]) -> dict:
    """Create a character object from the given definition file.

    The definition file should be an importable python file or a JSON
    file following one of the supported formats, filled with variables
    describing the character.

    This function also resolves any *parent_sheets* attributes in the
    given sheet, loading parent sheets and updating those attributes.

    Parameters
    ----------
    filename
      The path to the file that will be imported.

    Returns
    -------
    char_props
      Dictionary with the import character properties.

    """
    filename = Path(filename).resolve()
    # Parse the file name
    ext = filename.suffix
    try:
        reader = readers_by_extension[ext](filename=filename)
    except KeyError:
        raise ValueError(f"Character definition {filename} is not a known file type.")
    else:
        these_props = reader()
    # Resolve parent_sheets
    char_props = {}
    parent_sheets = these_props.pop("parent_sheets", [])
    for parent_sheet in parent_sheets:
        parent_sheet = (filename.parent / parent_sheet).resolve()
        if parent_sheet != filename:
            parent_props = read_sheet_file(parent_sheet)
            char_props.update(parent_props)
    char_props.update(these_props)
    # Remove imported dungeonsheets modules
    char_props.pop("import_homebrew", None)
    char_props.pop("mechanics", None)
    # Remove private variables (start with a '_')
    for attr in list(char_props.keys()):
        if attr[0] == "_":
            char_props.pop(attr)
    return char_props


class BaseCharacterReader:
    """Callable to parse a generic character file. Meant to be subclassed."""

    def __init__(self, filename: str):
        """

        Parameters
        ----------
        filename
          The path to the file that will be imported.

        """
        self.filename = filename

    def __call__(self, filename):
        """
        Parameters
        ----------
        filename
          The path to the file that will be imported.
        """
        raise NotImplementedError()


def json_character_reader(filename: str):
    """Factory to create a JSON reader of the correct sub-type."""
    # Try and extract the schema version to see if it's valid
    json_reader = Roll20CharacterReader(filename)
    try:
        json_reader.schema_version
    except KeyError:
        # Assume it's a foundry file
        json_reader = FoundryCharacterReader(filename)
    return json_reader


class JSONCharacterReader(BaseCharacterReader):
    """Callable to parse a JSON character file from Roll20 VVTES.

    The definition file should be a JSON file following one of the
    supported formats, filled with variables describing the character.

    """

    @lru_cache()
    def json_data(self):
        # Load the JSON data from disk
        with open(self.filename, mode="r") as fp:
            data = json.load(fp)
        return data

    def as_int(self, val):
        try:
            val = int(val)
        except ValueError:
            val = 0
        return val


class Roll20CharacterReader(JSONCharacterReader):
    @property
    def schema_version(self):
        return self.json_data()["schema_version"]

    def get_attrib(self, key, which="current", default=None):
        for obj in self.json_data()["attribs"]:
            if obj["name"] == key:
                val = obj[which]
                return val
        # No object was found
        if default is not None:
            return default
        else:
            raise KeyError(key)

    def has_skill_proficiency(self, key):
        false_profs = ["", "0"]
        return self.get_attrib(key=f"{key}_prof") not in false_profs

    def proficiencies(self, kind=None):
        """Iterator over the skills in which the character is proficient.

        *kind* can be one of "weapon", "language", or None (all
         proficiencies).

        """
        prof_re = re.compile("repeating_proficiencies_([-0-9a-zA-Z]+)_name")
        for obj in self.json_data()["attribs"]:
            match = prof_re.match(obj["name"])
            if match:
                prof_id = match.group(1)
                prof_type = self.get_attrib(
                    f"repeating_proficiencies_{prof_id}_prof_type"
                )
                if kind is None or prof_type == kind.upper():
                    yield self.get_attrib(
                        f"repeating_proficiencies_{prof_id}_name"
                    ).lower()

    def equipment(self, kind=None):
        """Iterator over items in the character's inventory."""
        prof_re = re.compile("repeating_inventory_([-0-9a-zA-Z]+)_itemname")
        for obj in self.json_data()["attribs"]:
            match = prof_re.match(obj["name"])
            if match:
                item_id = match.group(1)
                item_name = self.get_attrib(match.group(0))
                item_count = int(
                    self.get_attrib(
                        f"repeating_inventory_{item_id}_itemcount", default=1
                    )
                )
                # item_weight = self.get_attrib(
                #     f"repeating_inventory_{item_id}_itemweight", default=0
                # )
                item_str = item_name.lower().strip()
                if item_count > 1:
                    item_str += f" ({item_count})"
                yield item_str

    def weapons(self):
        """Iterator over the weapons the character is carrying in her inventory."""
        item_re = re.compile("repeating_attack_([-0-9a-zA-Z]+)_atkname")
        for obj in self.json_data()["attribs"]:
            match = item_re.match(obj["name"])
            if match:
                weapon_name = self.get_attrib(match.group()).lower()
                weapon_name = weapon_name.split("(")[0].strip()
                weapon_name = weapon_name.split(",")[0].strip()
                yield weapon_name

    def spells(self, prepared: bool = False):
        """Iterator over the spells the character knows.

        Parameters
        ==========
        prepared
          If true, only return prepared spells.

        """
        # "name": "repeating_spell-cantrip_-MEzYWPA5cUZYd4ZOvMS_spellname",
        prof_re = re.compile(
            "repeating_spell-(cantrip|[0-9]+)_([-0-9a-zA-Z]+)_spellname"
        )
        for obj in self.json_data()["attribs"]:
            match = prof_re.match(obj["name"])
            if match:
                level = match.group(1)
                spell_id = match.group(2)
                spell_name = self.get_attrib(
                    f"repeating_spell-{level}_{spell_id}_spellname"
                )
                is_prepared = self.as_int(
                    self.get_attrib(
                        f"repeating_spell-{level}_{spell_id}_spellprepared", default=0
                    )
                )
                if not prepared or is_prepared:
                    yield spell_name.lower()

    def __call__(self):
        """Create a character property dictionary from the JSON file."""
        # Verify the version compatibility
        if self.schema_version != 2:
            raise exceptions.JSONFormatError(
                "Cannot parse JSON schema version: %s" % self.schema_version
            )
        # Parse the json tree to get character properties
        char_props = {}
        char_props["name"] = self.json_data()["name"]
        char_props["level"] = self.as_int(self.get_attrib("base_level"))
        char_props["classes"] = [self.get_attrib("class")]
        char_props["background"] = self.get_attrib("background")
        char_props["race"] = self.get_attrib("subrace")
        char_props["alignment"] = self.get_attrib("alignment")
        char_props["xp"] = self.as_int(self.get_attrib("experience", default=0))
        # Attributes
        attribute_names = [
            "strength",
            "dexterity",
            "constitution",
            "intelligence",
            "wisdom",
            "charisma",
        ]
        for attr in attribute_names:
            char_props[attr] = self.as_int(self.get_attrib(f"{attr}_base"))
        # Skill proficiencies
        skill_names = [
            "acrobatics",
            "animal_handling",
            "arcana",
            "athletics",
            "deception",
            "history",
            "insight",
            "intimidation",
            "investigation",
            "medicine",
            "nature",
            "perception",
            "performance",
            "persuasion",
            "religion",
            "sleight_of_hand",
            "stealth",
            "survival",
        ]
        skill_profs = [
            skill for skill in skill_names if self.has_skill_proficiency(skill)
        ]
        char_props["skill_proficiencies"] = skill_profs
        # Other proficiencies
        char_props["weapon_proficiencies"] = self.proficiencies("weapon")
        char_props["languages"] = ", ".join(self.proficiencies("language"))
        # Tool proficiencies
        prof_re = re.compile("repeating_tool_([-0-9a-zA-Z]+)_toolname")
        tool_profs = []
        for obj in self.json_data()["attribs"]:
            match = prof_re.match(obj["name"])
            if match:
                tool_profs.append(self.get_attrib(match.group(0)))
        char_props["proficiencies_text"] = tool_profs
        # Combat stats
        char_props["hp_max"] = self.as_int(self.get_attrib("hp", which="max"))
        # Equipment
        char_props["cp"] = self.as_int(self.get_attrib("cp", default=0))
        char_props["sp"] = self.as_int(self.get_attrib("sp", default=0))
        char_props["ep"] = self.as_int(self.get_attrib("ep", default=0))
        char_props["gp"] = self.as_int(self.get_attrib("gp", default=0))
        char_props["pp"] = self.as_int(self.get_attrib("pp", default=0))
        char_props["weapons"] = self.weapons()
        char_props["equipment"] = ", ".join(self.equipment())
        # Personality, etc
        char_props["personality_traits"] = self.get_attrib("personality_traits").strip()
        char_props["flaws"] = self.get_attrib("flaws").strip()
        char_props["ideals"] = self.get_attrib("ideals").strip()
        char_props["bonds"] = self.get_attrib("bonds").strip()
        # Spells
        char_props["spells"] = self.spells()
        char_props["spells_prepared"] = self.spells(prepared=True)
        # Some unused values
        warn_msg = (
            "Importing the following traits from JSON is not yet supported: "
            "magic_items, armor, shield, attacks_and_spellcasting, "
            "infusions, wild_shapes"
        )
        warnings.warn(warn_msg)
        log.warning(warn_msg)
        char_props["magic_items"] = ()
        char_props["armor"] = ""
        char_props["shield"] = ""
        char_props["attacks_and_spellcasting"] = ""
        char_props["infusions"] = []
        char_props["wild_shapes"] = []
        return char_props


class FoundryCharacterReader(JSONCharacterReader):
    # List of weapons to ignore, only for class features that get added automatically
    _invalid_weapons = [
        "unarmed strike (monk)",
        "<no name>",
    ]

    def _skill_proficiency_value(self, key: str) -> float:
        proficiency_labels = {
            "acrobatics": "acr",
            "animal_handling": "ani",
            "arcana": "arc",
            "athletics": "ath",
            "deception": "dec",
            "history": "his",
            "insight": "ins",
            "intimidation": "itm",
            "investigation": "inv",
            "medicine": "med",
            "nature": "nat",
            "perception": "prc",
            "performance": "prf",
            "persuasion": "per",
            "religion": "rel",
            "sleight_of_hand": "slt",
            "stealth": "ste",
            "survival": "sur",
        }
        proficiency_value = float(
            self.json_data()["data"]["skills"][proficiency_labels[key]]["value"]
        )
        return proficiency_value

    def skill_proficiency(self, key: str) -> bool:
        return self._skill_proficiency_value(key) >= 1.0

    def skill_expertise(self, key: str) -> bool:
        return self._skill_proficiency_value(key) > 1.0

    def proficiencies(self, kind=None):
        """Iterator over the skills in which the character is proficient.

        *kind* can be one of "weapon", "language", or None (all
         proficiencies).

        """
        yield_weapons = kind is None or kind == "weapon"
        yield_languages = kind is None or kind == "language"
        # Weapon proficiencies
        if yield_weapons:
            weapon_prof = self.json_data()["data"]["traits"]["weaponProf"]
            # Simple and martial weapons
            if "sim" in weapon_prof["value"]:
                yield "simple weapons"
            if "mar" in weapon_prof["value"]:
                yield "martial weapons"
            # Extra weapons
            for weapon in weapon_prof["custom"].split(";"):
                yield weapon.strip()
        if yield_languages:
            # Languages
            lang_json = self.json_data()["data"]["traits"]["languages"]
            languages = lang_json["value"]
            languages.extend([s.strip() for s in lang_json["custom"].split(";")])
            yield from languages

    def weapons(self):
        """Iterator over the weapons the character is carrying in her inventory."""
        items = self.json_data()["items"]
        for item in items:
            is_valid_weapon = (
                item["type"] == "weapon"
                and item["name"].lower() not in self._invalid_weapons
            )
            if is_valid_weapon:
                yield item["name"].lower()

    def armor(self):
        items = self.json_data()["items"]
        armor_types = ["light", "medium", "heavy"]
        for item in items:
            if (
                item["type"] == "equipment"
                and item["data"]["armor"]["type"] in armor_types
            ):
                return item["name"].lower()

    def shield(self):
        items = self.json_data()["items"]
        for item in items:
            if (
                item["type"] == "equipment"
                and item["data"]["armor"]["type"] == "shield"
            ):
                return item["name"].lower()

    def equipment(self):
        """Iterator over the non-weapons the character is carrying in her
        inventory.

        """
        items = self.json_data()["items"]
        item_types = ["consumable", "equipment", "tool", "backpack", "loot"]
        for item in items:
            if item["type"] in item_types:
                item_name = item["name"]
                quantity = self.as_int(item["data"]["quantity"])
                if quantity != 1:
                    item_name += f"({quantity})"
                yield item_name.lower()

    def class_levels(self):
        for item in self.json_data()["items"]:
            if item["type"] == "class":
                yield (item["name"], item["data"]["levels"])

    def spells(self, prepared: bool = False):
        """Iterator over the spells the character knows.

        Parameters
        ==========
        prepared
          If true, only return prepared spells.

        """
        spells = (item for item in self.json_data()["items"] if item["type"] == "spell")
        if prepared:
            spells = (d for d in spells if d["data"]["preparation"]["prepared"])
        spell_names = (d["name"] for d in spells)
        yield from spell_names

    def __call__(self):
        """Create a character property dictionary from the JSON file."""
        # Parse the json tree to get character properties
        char_props = {}
        json_data = self.json_data()
        details = json_data["data"]["details"]
        char_props["name"] = json_data["name"]
        char_props["background"] = details["background"]
        char_props["race"] = details["race"]
        char_props["alignment"] = details["alignment"]
        char_props["xp"] = self.as_int(details["xp"]["value"])
        classes, levels = zip(*self.class_levels())
        char_props["levels"] = list(levels)
        char_props["classes"] = list(classes)
        # Attributes
        attribute_names = {
            "str": "strength",
            "dex": "dexterity",
            "con": "constitution",
            "int": "intelligence",
            "wis": "wisdom",
            "cha": "charisma",
        }
        abilities = self.json_data()["data"]["abilities"]
        save_proficiences = []
        for abbr, attr in attribute_names.items():
            char_props[attr] = self.as_int(abilities[abbr]["value"])
            # Check proficiency
            is_proficient = bool(abilities[abbr]["proficient"])
            if is_proficient:
                save_proficiences.append(attr)
        char_props["saving_throw_proficiencies"] = save_proficiences
        # Skill proficiencies
        skill_names = [
            "acrobatics",
            "animal_handling",
            "arcana",
            "athletics",
            "deception",
            "history",
            "insight",
            "intimidation",
            "investigation",
            "medicine",
            "nature",
            "perception",
            "performance",
            "persuasion",
            "religion",
            "sleight_of_hand",
            "stealth",
            "survival",
        ]
        skill_profs = [skill for skill in skill_names if self.skill_proficiency(skill)]
        char_props["skill_proficiencies"] = skill_profs
        skill_expertise = [
            skill for skill in skill_names if self.skill_expertise(skill)
        ]
        char_props["skill_expertise"] = skill_expertise
        # Other proficiencies
        char_props["weapon_proficiencies"] = self.proficiencies("weapon")
        char_props["languages"] = ", ".join(self.proficiencies("language"))
        # Tool proficiencies
        tool_labels = {
            "art": "artisan's tools",
            "disg": "disguise kit",
            "forg": "forger's kit",
            "game": "gaming set",
            "herb": "herbalism kit",
            "music": "musical instrument",
            "navg": "navigator's tools",
            "pois": "poisoner's kit",
            "thief": "thieves' tools",
            "vehicle": "vehicle (land or water)",
        }
        tool_profs = json_data["data"]["traits"]["toolProf"]["value"]
        tool_profs = [tool_labels[prof] for prof in tool_profs]
        custom_tool_profs = json_data["data"]["traits"]["toolProf"]["custom"]
        tool_profs.extend([s.strip() for s in custom_tool_profs.split(";")])
        char_props["proficiencies_text"] = tool_profs
        # Combat stats
        char_props["hp_max"] = self.as_int(json_data["data"]["attributes"]["hp"]["max"])
        # Equipment
        currency = json_data["data"]["currency"]
        char_props["cp"] = currency["cp"]
        char_props["sp"] = currency["sp"]
        char_props["ep"] = currency["ep"]
        char_props["gp"] = currency["gp"]
        char_props["pp"] = currency["pp"]
        char_props["weapons"] = list(self.weapons())
        char_props["equipment"] = ", ".join(self.equipment())
        char_props["armor"] = self.armor()
        char_props["shield"] = self.shield()
        # Personality, etc
        char_props["personality_traits"] = details["trait"].strip()
        char_props["flaws"] = details["flaw"].strip()
        char_props["ideals"] = details["ideal"].strip()
        char_props["bonds"] = details["bond"].strip()
        # Spells
        char_props["spells"] = self.spells()
        char_props["spells_prepared"] = self.spells(prepared=True)
        # Some unused values
        warn_msg = (
            "Importing the following traits from JSON is not yet supported: "
            "magic_items, attacks_and_spellcasting, "
            "infusions, wild_shapes."
        )
        warnings.warn(warn_msg)
        log.warning(warn_msg)
        char_props["magic_items"] = ()
        char_props["attacks_and_spellcasting"] = ""
        char_props["infusions"] = []
        char_props["wild_shapes"] = []
        return char_props


class PythonCharacterReader(BaseCharacterReader):
    def __call__(self):
        """Create a character object from the given definition file.

        The definition file should be an importable python file, filled
        with variables describing the character.

        Parameters
        ----------
        filename
          The path to the file that will be imported.

        """
        filename = self.filename
        # Check if this file contains the version string
        version_re = re.compile(r'dungeonsheets_version\s*=\s*[\'"]([0-9.]+)[\'"]')
        with open(filename, mode="r") as f:
            version = None
            for line in f:
                match = version_re.match(line)
                if match:
                    version = match.group(1)
                    break
            if version is None:
                # Not a valid DND character file
                raise exceptions.CharacterFileFormatError(
                    f"No ``dungeonsheets_version = `` entry in `{filename}`."
                )
        # Import the module to extract the information
        spec = importlib.util.spec_from_file_location("module", filename)
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)
        # Prepare a list of properties for this character
        char_props = {}
        for prop_name in dir(module):
            if prop_name[0:2] != "__":
                char_props[prop_name] = getattr(module, prop_name)
        return char_props


readers_by_extension = {
    ".py": PythonCharacterReader,
    ".json": json_character_reader,
}



from dungeonsheets import spells
from dungeonsheets.features.features import Feature


class _SpecialistSpells(Feature):
    """Starting at 3rd level, you always have certain spells pre­pared after
    you reach particular levels in this class, as shown in the Specialization
    Spells table. These spells count as artificer spells for you, but they
    don't count against the number of artificer spells you prepare.
    """

    _name = "Select One"
    source = "Artificer"
    _spells = {3: [], 5: [], 9: [], 13: [], 17: []}
    spells_known = []
    spells_prepared = []

    @property
    def name(self):
        return "{:s} Spells".format(self._name)

    def __init__(self, owner=None):
        if owner is not None:
            level = owner.Artificer.level
            for lvl, spl in self._spells.items():
                if level >= lvl:
                    self.spells_known.extend(spl)
                    self.spells_prepared.extend(spl)
        super().__init__(owner=owner)


# Alchemist
class ArtificerSpellcasting(Feature):
    """You have studied the workings of magic and how to channel it through
    objects. As a result, you have gained the ability to cast spells. To
    observers, you don't appear to be casting spells in a conventional way; you
    look as if you're producing wonders using mundane items or out­landish
    inventions.

    **Tools Required** You produce your artificer spell effects through your
    tools. You must have a spellcasting focus -- specifically thieves' tools or
    some kind of artisan's tool -- in hand when you cast any spell with this
    Spellcasting feature. You must be proficient with the tool to use it in
    this way. See chapter 5, "Equipment," in the Player's Handbook for
    descriptions of these tools.

    After you gain the Infuse Item feature at 2nd level, you can also use any
    item bearing one of your infusions as a spellcasting focus.
    """

    name = "Spellcasting"
    source = "Artificer"


class ArtificerRitualCasting(Feature):
    """You can cast an artificer spell as a ritual if that spell has the ritual
    tag and you have the spell prepared.
    """

    name = "Ritual Casting"
    source = "Artificer"


class FirearmProficiency(Feature):
    """The secrets of creating and operating gunpowder weapons have been
    discovered in various corners of the D&D multiverse. If your Dungeon Master
    uses the rules on firearms in chapter 9 of the Dungeon Master's Guide and
    your artificer has been exposed to the operation of such weapons, your
    artificer is proficient with them.
    """

    name = "Optional Rule: Firearm Proficiency"
    source = "Artificer"


class MagicalTinkering(Feature):
    """At 1st level, you learn how to invest a spark of magic into mundane
    objects. To use this ability, you must have tinker's tools or
    other artisan's tools in hand. You then touch a Tiny nonmagical
    object as an action and give it one of the following magical
    properties of your choice:

    - The object sheds bright light in a 5-foot radius and dim light
      for an additional 5 feet.
    - Whenever tapped by a creature, the object emits a recorded
      message that can be heard up to 10 feet away. You utter the
      message when you bestow this property on the object, and the
      recording can be no more than 6 seconds long.
    - The object continuously emits your choice of an odor or a
      nonverbal sound (wind, waves, chirping, or the like). The chosen
      phenomenon is perceivable up to 10 feet away.
    - A static visual effect appears on one of the object's
      surfaces. This effect can be a picture, up to 25 words of text,
      lines and shapes, or a mixture of these elements, as you
      like. The chosen property lasts indefinitely. As an action, you
      can touch the object and end the property early. You can bestow
      magic on multiple objects, touching one object each time you use
      this feature, though a single object can only bear one property
      at a time. The maximum number of objects you can affect with
      this feature at one time is equal to your Intelligence modifier
      (minimum of one object). If you try to exceed your maximum, the
      oldest property immediately ends, and then the new property
      applies.

    """

    name = "Magical Tinkering"
    source = "Artificer"


class InfuseItem(Feature):
    """At 2nd level, you gain the ability to imbue mundane items with certain
    magical infusions. The magic items you create with this feature are
    effectively prototypes of permanent items.

    Infusions known
      When you gain this feature, pick four artificer infusions to
      learn, choosing from the "Artificer Infusions" section at the
      end of the class's description. You learn additional infusions
      of your choice when you reach certain levels in this class, as
      shown in the Infusions Known column of the Artificer
      table. Whenever you gain a level in this class, you can re­place
      one of the artificer infusions you learned with a new one.
    Infusing an item
      Whenever you finish a long rest, you can touch a non­magical
      object and imbue it with one of your artificer in­fusions,
      turning it into a magic item. An infusion works on only certain
      kinds of objects, as specified in the infusion's description. If
      the item requires attunement, you can attune yourself to it the
      instant you infuse the item. If you decide to attune to the item
      later, you must do so using the normal process for attunement
      (see "Attunement" in chapter 7 of the Dungeon Master's Guide).
      Your infusion remains in an item indefinitely, but when you die,
      the infusion vanishes after a number of days have passed equal
      to your Intelligence modifier (minimum of 1 day). The infusion
      also vanishes if you give up your knowledge of the infusion for
      another one. You can infuse more than one nonmagical object at
      the end of a long rest; the maximum number of objects appears in
      the Infused Items column of the Artificer table. You must touch
      each of the objects, and each of your infusions can be in only
      one object at a time. Moreover, no object can bear more than one
      of your infusions at a time. If you try to exceed your maximum
      number of in­fusions, the oldest infusion immediately ends, and
      then the new infusion applies.

    """

    _name = "Infuse Item"
    source = "Artificer"
    _infusions = {
        # level: (infusions known, infused items)
        2: (4, 2),
        3: (4, 2),
        4: (4, 2),
        5: (4, 2),
        6: (6, 3),
        7: (6, 3),
        8: (6, 3),
        9: (6, 3),
        10: (8, 4),
        11: (8, 4),
        12: (8, 4),
        13: (8, 4),
        14: (10, 5),
        15: (10, 5),
        16: (10, 5),
        17: (10, 5),
        18: (12, 6),
        19: (12, 6),
        20: (12, 6),
    }

    @property
    def name(self):
        known_infusions = self._infusions[self.owner.Artificer.level][0]
        infused_items = self._infusions[self.owner.Artificer.level][1]
        name_ext = " ({:d} Infusions Known, {:d} Infused Items)"
        return self._name + name_ext.format(known_infusions, infused_items)


class ArtificerSpecialist(Feature):
    """At 3rd level, you choose the type of specialist you are: Alchemist,
    Artillerist, or Battle Smith, each of which is detailed at the end of the
    class's description. Your choice grants you features at 5th level and again
    at 9th and 15th level.
    """

    name = "Artificer Specialist"
    source = "Artificer"


class TheRightToolForTheJob(Feature):
    """At 3rd level, you learn how to produce exactly the tool you need: with
    tinker's tools in hand, you can magically create one set of artisan's tools
    in an unoccupied space within 5 feet of you. This creation requires 1 hour
    of uninterrupted work, which can coincide with a short or long rest.
    Though the product of magic, the tools are nonmagical, and they vanish when
    you use this feature again.
    """

    name = "The Right Tool For The Job"
    source = "Artificer"


class ToolExpertise(Feature):
    """Starting at 6th level, your proficiency bonus is doubled for any ability
    check you make that uses your proficiency with a tool.
    """

    name = "Tool Expertise"
    source = "Artificer"


class FlashOfGenius(Feature):
    """Starting at 7th level, you gain the ability to come up with
    solutions under pressure. When you or another creature you can see
    within 30 feet of you makes an ability check or a saving throw,
    you can use your reaction to add your Intelligence modifier to the
    roll.

    You can use this feature a number of times equal to your
    Intelligence modifier (minimum of once). You regain all expended
    uses when you finish a long rest.

    """

    name = "Flash Of Genius"
    source = "Artificer"


class MagicItemAdept(Feature):
    """When you reach 10th level, you achieve a profound understanding of how
    to use and make magic items:

    - You can attune to up to four magic items at once.
    - If you craft a magic item with a rarity of common or uncommon,
      it takes you a quarter of the normal time, and it costs you half
      as much of the usual gold.

    """

    name = "Magic Item Adept"
    source = "Artificer"


class SpellStoringItem(Feature):
    """At 11th level, you learn how to store a spell in an
    object. Whenever you finish a long rest, you can touch one simple
    or martial weapon or one item that you can use as a spellcasting
    focus, and you store a spell in it, choosing a 1st- or 2nd-level
    spell from the artificer spell list that requires 1 action to cast
    (you needn't have it prepared).

    While holding the object, a creature can take an action to produce
    the spell's effect from it, using your spellcasting ability
    modifier. If the spell requires concentration, the creature must
    concentrate. The spell stays in the object until it's been used a
    number of times equal to twice your Intelligence modifier (minimum
    of twice) or until you use this fe ature again to store a spell in
    an object.
    """

    name = "Spell-Storing Item"
    source = "Artificer"


class MagicItemSavant(Feature):
    """At 14th level, your skill with magic items deepens more:

    - You can attune to up to five magic items at once.
    - You ignore all class, race, spell, and level requirements on
      attuning to or using a magic item.

    """

    name = "Magic Item Savant"
    source = "Artificer"


class MagicItemMaster(Feature):
    """Starting at 18th level, you can attune to up to six magic items at
    once.
    """

    name = "Magic Item Master"
    source = "Artificer"


class SoulOfArtifice(Feature):
    """At 20th level, you develop a mystical connection to your magic items,
    which you can draw on for protection:

    - You gain a +1 bonus to all saving throws per magic item you are currently
      attuned to.
    - If you're reduced to 0 hit points but not killed out­right, you
      can use your reaction to end one of your artificer infusions,
      causing you to drop to 1 hit point instead of 0.

    """

    name = "Soul of Artifice"
    source = "Artificer"


# Alchemist
class AlchemistToolProficiency(Feature):
    """When you adopt this specialization at 3rd level, you gain proficiency
    with alchemist's supplies. If you already have this proficiency, you gain
    proficiency with one other type of artisan's tools of your choice.
    """

    name = "Tool Proficiency"
    source = "Artificer (Alchemist)"


class AlchemistSpells(_SpecialistSpells):
    """Starting at 3rd level, you always have certain spells pre­pared after
    you reach particular levels in this class, as shown in the Alchemist Spells
    table. These spells count as artificer spells for you, but they don't count
    against the number of artificer spells you prepare.
    """

    _name = "Alchemist"
    _spells = {
        3: [spells.HealingWord, spells.RayOfSickness],
        5: [spells.FlamingSphere, spells.MelfsAcidArrow],
        9: [spells.GaseousForm, spells.MassHealingWord],
        13: [spells.Blight, spells.DeathWard],
        17: [spells.Cloudkill, spells.RaiseDead],
    }


class ExperimentalElixir(Feature):
    """Beginning at 3rd level, whenever you finish a long rest, you can
    magically produce an *experimental elixir* in an empty flask you touch.
    Roll on the Experimental Elixir table for the elixir's effect, which is
    triggered when someone drinks the elixir. As an action, a creature can
    drink the elixir or administer it to an incapacitated creature.

    Creating an *experimental elixir* requires you to have alchemist supplies
    on your person, and any elixir you create with this feature lasts until it
    is drunk or until the end of your next long rest.

    When you reach certain levels in this class, you can make more elixirs at
    the end of a long rest: two at 6th level and three at 15th level. Roll for
    each elixir's effect separately. Each elixir requires its own flask.

    You can create additional *experimental elixirs* by expending a spell slot
    of 1st level or higher for each one. When you do so, you use your action to
    create the elixir in an empty flask you touch, and you choose the elixir's
    effect from the Experimental Elixir table.

    **Experimental Elixir**

    roll d6

    **1 -- Healing.** The drinker regains a number of hit points equal to 2d4 +
    your Intelligence modifier.

    **2 -- Swiftness.** The drinker's walking speed increases by 10 feet for 1
    hour.

    **3 -- Resilience.** The drinker gains a +1 bonus to AC for 10 minutes.

    **4 -- Boldness.** The drinker can roll a d4 and add the num­ber rolled to
    every attack roll and saving throw they make for the next minute.

    **5 -- Flight.** The drinker gains a flying speed of 10 feet for 10
    minutes.

    **6 -- Transformation.** The drinker's body is transformed as if by the
    alter self spell. The drinker determines the transformation caused by the
    spell, the effects of which last for 10 minutes.
    """

    name = "Experimental Elixir"
    source = "Artificer (Alchemist)"


class AlchemicalSavant(Feature):
    """At 5th level, you develop masterful command of magical chemicals,
    enhancing the healing and damage you create through them. Whenever you
    cast a spell using your alchemist's supplies as the spellcasting focus, you
    gain a bonus to one roll of the spell. That roll must restore hit points or
    be a damage roll that deals acid, fire, necrotic, or poison damage, and the
    bonus equals your Intelligence modifier (minimum of +1).
    """

    name = "Alchemical Savant"
    source = "Artificer (Alchemist)"


class RestorativeReagents(Feature):
    """Starting at 9th level, you can incorporate restorative reagents
    into some of your works:

    - Whenever a creature drinks an experimental elixir you created,
      the creature gains temporary hit points equal to 2d6 + your
      Intelligence modifier (minimum of 1 temporary hit point).
    - You can cast lesser restoration without expending a spell slot
      and without preparing the spell, provided you use alchemist's
      supplies as the spellcasting focus. You can do so a number of
      times equal to your Intelligence modifier (minimum of once), and
      you regain all expended uses when you finish a long rest.

    """

    name = "Restorative Reagents"
    source = "Artificer (Alchemist)"


class ChemicalMastery(Feature):
    """By 15th level, you have been exposed to so many chemicals that they
    pose little risk to you, and you can use them to quickly end
    certain ailments:

    - You gain resistance to acid damage and poison damage, and you
      are immune to the poisoned condition.
    - You can cast greater restoration and heal without expending a
      spell slot, without preparing the spell, and without material
      components, provided you use alchemist's supplies as the
      spellcasting focus. Once you cast either spell with this
      feature, you can't cast that spell with it again until you
      finish a long rest.

    """

    name = "Chemical Mastery"
    source = "Artificer (Alchemist)"


# Artillerist
class ArtilleristSpells(_SpecialistSpells):
    """Starting at 3rd level, you always have certain spells prepared after
    you reach particular levels in this class, as shown in the Artillerist
    Spells table. These spells count as artificer spells for you, but they
    don't count against the number of artificer spells you prepare.
    """

    _name = "Artillerist"
    _spells = {
        3: [spells.Shield, spells.Thunderwave],
        5: [spells.ScorchingRay, spells.Shatter],
        9: [spells.Fireball, spells.WindWall],
        13: [spells.IceStorm, spells.WallOfFire],
        17: [spells.ConeOfCold, spells.WallOfForce],
    }


class ArtilleristToolProficiency(Feature):
    """When you adopt this specialization at 3rd level, you gain proficiency
    with woodcarver's tools. If you already have this proficiency, you gain
    proficiency with one other type of artisan's tools of your choice.
    """

    name = "Tool Proficiency"
    source = "Artificer (Artillerist)"


class EldritchCannon(Feature):
    """At 3rd level, you learn how to create a magical cannon. Using
    woodcarver's tools or smith's tools, you can take an action to magically
    create a Small or Tiny eldritch cannon in an unoccupied space on a
    horizontal surface within 5 feet of you. A Small eldritch cannon occupies
    its space, and a Tiny one can be held in one hand.

    Once you create a cannon, you can't do so again until you finish a long
    rest or until you expend a spell slot of 1st level or higher. You can have
    only one cannon at a time and can't create one while your cannon is
    present.

    The cannon is a magical object. Regardless of size, the cannon has an AC of
    18 and a number of hit points equal to five times your artificer level. It
    is immune to poison damage, psychic damage, and all conditions. If it is
    forced to make an ability check or a saving throw, treat all its ability
    scores as 10 (+O). If the *mending* spell is cast on it, it regains 2d6 hit
    points. It disappears if it is reduced to 0 hit points or after 1 hour. You
    can dismiss it early as an action.

    When you create the cannon, you determine its appearance and whether it
    has legs. You also decide which type it is, choosing from the options on
    the Eldritch Cannons table. On each of your turns, you can take a bonus
    action to cause the cannon to activate if you are within 60 feet of it. As
    part of the same bonus action, you can direct the cannon to walk or climb
    up to 15 feet to an unoccupied space, provided it has legs.

    **Eldritch Cannons**

    *Flamethrower*: The cannon exhales fire in an adjacent 15-foot cone that
    you designate. Each creature in that area must make a Dexterity saving
    throw against your spell save DC, taking 2d8 fire damage on a failed save
    or half as much damage on a successful one. The fire ignites any flammable
    objects in the area that aren't being worn or carried.

    *Force Ballista*: Make a ranged spell attack, originating from the cannon,
    at one creature or object within 120 feet of it. On a hit, the target takes
    2d8 force damage, and if the target is a creature, it is pushed up to 5
    feet away from the cannon.

    *Protector*: The cannon emits a burst of positive energy
    that grants itself and each creature of your choice within 10 feet of it a
    number of temporary hit points equal to 1d8 + your Intelligence modifier
    (minimum of +1)
    """

    name = "Eldritch Cannon"
    source = "Artificer (Artillerist)"


class ArcaneFirearm(Feature):
    """At 5th level, you know how to turn a wand, staff, or rod into an
    arcane firearm, a conduit for your destructive spells. When you
    finish a long rest, you can use wood­carver's tools to carve
    special sigils into a wand, staff, or rod and thereby turn it into
    your arcane firearm. The sigils disappear from the object if you
    later carve them on a different item. The sigils otherwise last
    indefinitely.

    You can use your arcane firearm as a spellcasting focus for your
    artificer spells. When you cast an artificer spell through the
    firearm, roll a d8, and you gain a bonus to one of the spell's
    damage rolls equal to the number rolled.

    """

    name = "Arcane Firearm"
    source = "Artificer (Artillerist)"


class ExplosiveCannon(Feature):
    """Starting at 9th level, every eldritch cannon you create is more
    destructive:

    - The cannon's damage rolls all increase by 1d8.
    - As an action, you can command the cannon to detonate if you are
      within 60 feet of it. Doing so destroys the cannon and forces
      each creature within 20 feet of it to make a Dexterity saving
      throw against your spell save DC, taking 3d8 force damage on a
      failed save or half as much damage on a successful one.

    """

    name = "Explosive Cannon"
    source = "Artificer (Artillerist)"


class FortifiedPosition(Feature):
    """Starting at 15th level, you're a master at forming well-defended
    emplacements using Eldritch Cannon:

    - You and your allies have half cover while within 10 feet of a
      cannon you create with Eldritch Cannon, as a result of a
      shimmering field of magical protection that the cannon emits.
    - You can now have two cannons at the same time. You can create
      two with the same action (but not the same spell slot), and you
      can activate both of them with the same bonus action. You
      determine whether the cannons are identical to each other or
      different. You can't create a third cannon while you have two.

    """

    name = "Fortified Position"
    source = "Artificer (Artillerist)"


# Battle Smith
class BattleSmithSpells(_SpecialistSpells):
    """Starting at 3rd level, you always have certain spells prepared
    after you reach particular levels in this class, as shown in the
    Battle Smith Spells table. These spells count as artificer spells
    for you, but they don't count against the number of artificer
    spells you prepare.

    """

    _name = "Battle Smith"
    _spells = {
        3: [spells.Heroism, spells.Shield],
        5: [spells.BrandingSmite, spells.WardingBond],
        9: [spells.AuraOfVitality, spells.ConjureBarrage],
        13: [spells.AuraOfPurity, spells.FireShield],
        17: [spells.BanishingSmite, spells.MassCureWounds],
    }


class BattleSmithToolProficiency(Feature):
    """When you adopt this specialization at 3rd level, you gain
    proficiency with smith's tools. If you already have this
    proficiency, you gain proficiency with one other type of artisan's
    tools of your choice.

    """

    name = "Tool Proficiency"
    source = "Artificer (Battle Smith)"


class BattleReady(Feature):
    """When you reach 3rd level, your combat training and your experiments with
    magic have paid off in two ways:

    - You gain proficiency with martial weapons.
    - When you attack with a magic weapon, you can use your
      Intelligence modifier, instead of Strength or Dexterity
      modifier, for the attack and damage rolls.

    """

    name = "Battle Ready"
    source = "Artificer (Battle Smith)"


class SteelDefender(Feature):
    """By 3rd level, your tinkering has borne you a faithful companion, a
    steel defender. It is friendly to you and your companions, and it
    obeys your commands. See this creature's game statistics in the
    steel defender stat block. You determine the creature's appearance
    and whether it has two legs or four; your choice has no effect on
    its game statistics.

    In combat, the steel defender shares your initiative count, but it
    takes its turn immediately after yours. It can move and use its
    reaction on its own, but the only action it takes on its turn is
    the Dodge action, unless you take a bonus action on your turn to
    command it to take one of the actions in its stat block or the
    Dash, Disengage, Help, Hide, or Search action.

    If the *mending* spell is cast on it, it regains 2d6 hit
    points. If it has died within the last hour, you can use your
    smith's tools as an action to revive it, provided you are within 5
    feet of it and you expend a spell slot of 1st level or higher. The
    steel defender returns to life after 1 minute with all its hit
    points restored.

    At the end of a long rest, you can create a new steel defender if
    you have your smith's tools with you. If you already have a steel
    defender from this feature, the first one immediately perishes.

    """

    name = "Steel Defender"
    source = "Artificer (Battle Smith)"


class ExtraAttackBattleSmith(Feature):
    """Starting at 5th level, you can attack twice, rather than once,
    whenever you take the Attack action on your turn.

    """

    name = "Extra Attack"
    source = "Artificer (Battle Smith)"


class ArcaneJolt(Feature):
    """At 9th level, you learn new ways to channel arcane energy to harm
    or heal. When either you hit a target with a magic weapon attack
    or your steel defender hits a target, you can channel magical
    energy through the strike to create one of the following effects:

    - The target takes an extra 2d6 force damage.
    - Choose one creature or object you can see within 30 feet of the
      target.  Healing energy flows into the chosen recipient,
      restoring 2d6 hit points to it. You can use this energy a number
      of times equal to your Intelligence modifier (minimum of once),
      but you can do so no more than once on a turn. You regain all
      expended uses when you finish a long rest

    """

    name = "Arcane Jolt"
    source = "Artificer (Battle Smith)"


class ImprovedDefender(Feature):
    """At 15th level, your Arcane jolt and steel defender be­come more
    powerful:

    - The extra damage and the healing of your Arcane jolt both
      increase to 4d6.
    - Your steel defender gains a +2 bonus to Armor Class.
    - Whenever your steel defender uses its Deflect Attack, the
      attacker takes force damage equal to 1d4 +your Intelligence
      modifier.

    """

    name = "Improved Defender"
    source = "Artificer (Battle Smith)"



class CharacterFileFormatError(ValueError):
    """The given file is not a valid Dungeons and Dragons file."""


class DiceError(ValueError):
    """Improper formatting for a dice string."""


class LatexError(OSError):
    """PDFLatex did not execute correctly."""


class LatexNotFoundError(LatexError):
    """PDFLatex did not execute correctly."""


class MonsterError(AttributeError):
    """Error retriving or using a D&D Monster."""


class JSONFormatError(RuntimeError):
    """The JSON file doesn't conform to the understood formats."""


class UnknownFileType(RuntimeError):
    """The input file does not match one of the known formats."""


class UnknownOutputFormat(RuntimeError):
    """The output format requested is not one of the known outputs."""


class ContentNotFound(ValueError):
    """The requested content could not be resolved."""

class AmbiguousContent(ValueError):
    """Multiple valid content entries were found."""



from dungeonsheets.spells.spells import Spell


class IceKnife(Spell):
    """(a drop of water or piece of ice)
    You create a shard of ice and fling it at one
    creature within range. Make a ranged spell attack against the target. On a hit,
    the target takes 1d10 piercing damage. Hit or miss, the shard then explodes. The
    target and each creature within 5 feet of the point where the ice exploded must
    succeed on a Dexterity saving throw or take 2d6 cold damage.
    At Higher Levels.
    When you cast this spell using a spell slot of 2nd level or higher, the cold
    damage increases by 1d6 for each slot level above 1st.
    """

    name = "Ice Knife"
    level = 1
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("S", "M")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Druid", "Sorcerer", "Wizard")


class IceStorm(Spell):
    """A hail of rock-hard ice pounds to the ground in a 20-foot-radius, 40-foot-high
    cylinder centered on a point within range.
    Each creature in the cylinder must
    make a Dexterity saving throw. A creature takes 2d8 bludgeoning damage and 4d6
    cold damage on a failed save, or half as much damage on a successful one.


    Hailstones turn the storm's area of effect into difficult terrain until the end
    of your next turn.

    **At Higher Levels:** When you cast this spell using a spell
    slot of 5th level or higher, the bludgeoning damage increases by 1d8 for each
    slot level above 4th.
    """

    name = "Ice Storm"
    level = 4
    casting_time = "1 action"
    casting_range = "300 feet"
    components = ("V", "S", "M")
    materials = "A pinch of dust and a few drops of water"
    duration = "Instantaneous"
    ritual = False
    magic_school = "Evocation"
    classes = ("Druid", "Sorcerer", "Wizard")


class Identify(Spell):
    """You choose one object that you must touch throughout the casting of the spell.
    If it is a magic item or some other magic-imbued object, you learn its
    properties and how to use them, whether it requires attunement to use, and how
    many charges it has, if any. You learn whether any spells are affecting the item
    and what they are. If the item was created by a spell, you learn which spell
    created it.

    If you instead touch a creature throughout the casting, you learn
    what spells, if any, are currently affecting it.
    """

    name = "Identify"
    level = 1
    casting_time = "1 minute"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "A pearl worth at least 100 gp and an owl feather"
    duration = "Instantaneous"
    ritual = True
    magic_school = "Divination"
    classes = ("Bard", "Wizard")


class IdInsinuation(Spell):
    """You unleash a torrent of conflicting desires in the mind of one creature you can
    see within range, impairing its ability to make decisions. The target must succeed
    on a Wisdom saving throw or be incapacitated. At the end of each of its turns, it
    takes 1d12 psychic damage, and it can then make another Wisdom saving throw. On a
    success, the spell ends on the target.
    """

    name = "Id Insinuation"
    level = 1
    casting_time = "1 action"
    casting_range = "60ft"
    components = ("V", "S")
    duration = "1 minute"
    magic_school = "Enchantment"
    classes = "Sorcerer"


class IllusoryDragon(Spell):
    """By gathering threads of shadow material from the Shadowfell, you create a Huge
    shadowy dragon in an unoccupied space that you can see within range. The
    illusion lasts for the spell's duration and occupies its space, as if it were a
    creature.
    When the illusion appears, any of your enemies that can see it must
    succeed on a Wisdom saving throw or become frightened of it for 1 minute. If a
    frightened creature ends its turn in a location where it doesn't have line of
    sight to the illusion, it can repeat the saving throw, ending the effect on
    itself on a success.
    As a bonus action on your turn, you can move the illusion
    up to 60 feet. At any point during its movement, you can cause it to exhale a
    blast of energy in a 60-foot cone originating from its space. When you create
    the dragon, choose a damage type: acid, cold, fire, lightning, necrotic, or
    poison. Each creature in the cone must make an Intelligence saving throw, taking
    '7d6 damage of the
    chosen damage type on a failed save, or half as much damage
    on a successful one.
    The illusion is tangible because of the shadow stuff used
    to create it, but attacks miss it automatically. it succeeds on all saving
    throws, and it is immune to all damage and conditions. A creature that uses an
    action to examine the dragon can determine that it is an illusion by succeeding
    on an Intelligence (Investigation) check against your spell save DC. If a
    creature discerns the illusion for what it is, the creature can see through it
    and has advantage on saving throws against its breath.
    """

    name = "Illusory Dragon"
    level = 8
    casting_time = "1 action"
    casting_range = "120 feet"
    components = ("S",)
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Illusion"
    classes = ("Wizard",)


class IllusoryScript(Spell):
    """You write on parchment, paper, or some other suitable writing material and imbue
    it with a potent illusion that lasts for the duration.

    To you and any
    creatures you designate when you cast the spell, the writing appears normal,
    written in your hand, and conveys whatever meaning you intended when you wrote
    the text. To all others, the writing appears as if it were written in an unknown
    or magical script that is unintelligible. Alternatively, you can cause the
    writing to appear to be an entirely different message, written in a different
    hand and language, though the language must be one you know.

    Should the spell
    be dispelled, the original script and the illusion both disappear.
    A creature
    with truesight can read the hidden message.
    """

    name = "Illusory Script"
    level = 1
    casting_time = "1 minute"
    casting_range = "Touch"
    components = ("S", "M")
    materials = "A lead-based ink worth at least 10 gp, which the spell consumes"
    duration = "10 days"
    ritual = True
    magic_school = "Illusion"
    classes = ("Bard", "Warlock", "Wizard")


class Immolation(Spell):
    """Flames wreathe one creature you can see within range. The target must make a
    Dexterity saving throw. It takes 8d6 fire damage on a failed save, or half as
    much damage on a successful one. On a failed save, the target also burns for the
    spell's duration. The burning target sheds bright light in a 30-foot radius and
    dim light for an additional 30 feet. At the end of each of its turns, the
    target repeats the saving throw. It takes 4d6 fire damage on a failed save, and
    the spell ends on a successful one. These magical flames can't be extinguished
    by nonmagical means.
    If damage from this spell kills a target, the target is
    turned to ash.
    """

    name = "Immolation"
    level = 5
    casting_time = "1 action"
    casting_range = "90 feet"
    components = ("V",)
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Evocation"
    classes = ("Sorcerer", "Wizard")


class Imprisonment(Spell):
    """You create a magical restraint to hold a creature that you can see within range.

    The target must succeed on a Wisdom saving throw or be bound by the spell; if
    it succeeds, it is immune to this spell if you cast it again. While affected by
    this spell, the creature doesn't need to breathe, eat, or drink, and it doesn't
    age. Divination spells can't locate or perceive the target.

    When you cast the
    spell, you choose one of the following forms of imprisonment.

    Burial
    The
    target is entombed far beneath the earth in a sphere of magical force that is
    just large enough to contain the target. Nothing can pass through the
    sphere, nor can any creature teleport or use planar travel to get into or out of
    it.
    The special component for this version of the spell is a small mithral orb.


    Chaining
    Heavy chains, firmly rooted in the ground, hold the target in place.
    The target is restrained until the spell ends, and it can't move or be moved by
    any means until then.
    The special component for this version of the spell is
    a fine chain of precious metal.

    Hedged Prison
    The spell transports the target
    into a tiny demiplane that is warded against teleportation and planar travel.
    The demiplane can be a labyrinth, a cage, a tower, or any similar confined
    structure or area of your choice.
    The special component for this version of the
    spell is a miniature representation of the prison made from jade.

    Minimus
    Containment
    The target shrinks to a height of 1 inch and is imprisoned inside a
    gemstone or similarobject. Light can pass through the gemstone
    normally (allowing the target to see out and other creatures to see in), but
    nothing else can pass through, even by means of teleportation or planar travel.
    The gemstone can't be cut or broken while the spell remains in effect.
    The
    special component for this version of the spell is a large, transparent
    gemstone, such as a corundum, diamond, or ruby.

    Slumber
    The target falls asleep
    and can't be awoken.
    The special component for this version of the
    spell consists of rare soporific herbs.

    Ending the Spell
    During the casting of
    the spell, in any of its versions, you can specify a condition that will cause
    the spell to end and release the target. The condition can be as specific or as
    elaborate as you choose, but the DM must agree that the condition is reasonable
    and has a likelihood of coming to pass. The conditions can be based on a
    creature's name, identity, or deity but otherwise must be based on
    observable actions or qualities and not based on intangibles such as level,
    class, or hit points.

    A dispel magic spell can end the spell only if it is
    cast as a 9th-level spell, targeting either the prison or the special component
    used to create it.

    You can use a particular special component to create only one
    prison at a time. If you cast the spell again using the same
    component, the target of the first casting is immediately freed
    from its binding.

    """

    name = "Imprisonment"
    level = 9
    casting_time = "1 minute"
    casting_range = "30 feet"
    components = ("V", "S", "M")
    materials = (
        "A vellum depiction or a carved statuette in the likeness of the target, and a"
        " special component that varies according to the version of the spell you"
        " choose, worth at least 500 gp per hit die of the target"
    )
    duration = "Until dispelled"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Warlock", "Wizard")


class IncendiaryCloud(Spell):
    """A swirling cloud of smoke shot through with white-hot embers appears in a
    20-foot-radius sphere centered on a point within range.
    The cloud spreads around
    corners and is heavily obscured. It lasts for the duration or until a wind of
    moderate or greater speed (at least 10 miles per hour) disperses it.

    When the
    cloud appears, each creature in it must make a Dexterity saving throw. A
    creature takes 10d8 fire damage on a failed save, or half as much damage on a
    successful one. A creature must also make this saving throw when it enters the
    spell's area for the first time on a turn or ends its turn there.

    The cloud
    moves 10 feet directly away from you in a direction that you choose at the start
    of each of your turns.
    """

    name = "Incendiary Cloud"
    level = 8
    casting_time = "1 action"
    casting_range = "150 feet"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 1 minute"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Sorcerer", "Wizard")


class InfernalCalling(Spell):
    """Uttering a dark incantation, you summon a devil from the Nine Hells. You choose
    the devil's type, which must be one of challenge rating 6 or lower, such as a
    barbed devil or a bearded devil. The devil appears in an unoccupied space that
    you can see within range. The devil disappears when it drops to 0 hit points or
    when the spell ends.
    The devil is unfriendly toward you and your companions.
    Roll initiative for the devil, which has its own turns. It is under the Dungeon
    Master's control and acts according to its nature on each of its turns, which
    might result in its attacking you if it thinks it can prevail, or trying to
    tempt you to undertake an evil act in exchange for limited service. The DM has
    the creature's statistics.
    On each of your turns, you can try to issue a verbal
    command to the devil (no action required by you). It obeys the command if the
    likely outcome is in accordance with its desires, especially if the result would
    draw you toward evil. Otherwise, you must make a Charisma (Deception,
    Intimidation, or Persuasion) check contested by its Wisdom (Insight) check. You
    make the check with advantage if you say the devil's true name. Ifyour check
    fails, the devil becomes immune to your verbal commands for the duration of the
    spell, though it can still carry out your commands if it chooses. If your check
    succeeds, the devil carries out your command- such as "attack my enemies,"
    "explore the room ahead," or "bear this message to the queen"-until it completes
    the activity, at which point it returns to you to report having done so.
    If
    your concentration ends before the spell reaches its full duration, the devil
    doesn‘t disappear if it has become immune to your verbal commands. Instead, it
    acts in whatever manner it chooses for 3d6 minutes, and then it disappears.
    If
    you possess an individual devil's talisman, you can summon that devil if it is
    of the appropriate challenge
    rating plus 1, and it obeys all your commands, with
    no Charisma checks required.

    **At Higher Levels:** When you cast this spell using
    a spell slot of 6th level or higher, the challenge rating increases by 1 for
    each slot level above 5th.
    """

    name = "Infernal Calling"
    level = 5
    casting_time = "1 minute"
    casting_range = "90 feet"
    components = ("V", "S", "M")
    materials = "A ruby worth at least 999 gp"
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Warlock", "Wizard")


class Infestation(Spell):
    """You cause a cloud of mites, fleas, and other parasites to appear momentarily on
    one creature you can see within range. The target must succeed on a Constitution
    saving throw, or it takes 1d6 poison damage and moves 5 feet in a random
    direction if it can move and its speed is at least 5 feet. Roll a d4 for the
    direction: 1., north; 2, south; 3, east; or 4, west. This movement doesn't
    provoke opportunity attacks, and if the direction rolled is blocked, the target
    doesn't move.
    The spell's damage increases by 1d6 when you reach 5th level
    (2d6), 11th level (3d6), and 17th level (4d6).
    """

    name = "Infestation"
    level = 0
    casting_time = "1 action"
    casting_range = "30 feet"
    components = ("V", "S", "M")
    materials = "A living flea"
    duration = "Instantaneous"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Druid", "Sorcerer", "Warlock", "Wizard")


class InflictWounds(Spell):
    """Make a melee spell attack against a creature you can reach. On a hit, the target
    takes 3d10 necrotic damage.

    **At Higher Levels:** When you cast this spell using a
    spell slot of 2nd level or higher, the damage increases by 1d10 for each slot
    level above 1st.
    """

    name = "Inflict Wounds"
    level = 1
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S")
    materials = ""
    duration = "Instantaneous"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Cleric",)


class InsectPlague(Spell):
    """Swarming, biting locusts fill a 20-foot-radius sphere centered on a point you
    choose within range. The sphere spreads around corners. The sphere remains for
    the duration, and its area is lightly obscured. The sphere's area is difficult
    terrain.

    When the area appears, each creature in it must make a Constitution
    saving throw. A creature takes 4d10 piercing damage on a failed save, or half as
    much damage on a successful one. A creature must also make this saving throw
    when it enters the spell's area for the first time on a turn or ends its turn
    there.

    **At Higher Levels:** When you cast this spell using a spell slot of 6th
    level or higher, the damage increases by 1d10 for each slot level above 5th.
    """

    name = "Insect Plague"
    level = 5
    casting_time = "1 action"
    casting_range = "300 feet"
    components = ("V", "S", "M")
    materials = "A few grains of sugar, some kernels of grain, and a smear of fat"
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Conjuration"
    classes = ("Cleric", "Druid", "Sorcerer")


class InvestitureOfFlame(Spell):
    """Flames race across your body, shedding bright light in a 30-foot radius and dim
    light for an additional 30 feet for the spell's duration. The flames don't harm
    you. Until the spell ends, you gain the following benefits:
    - You are immune to
    fire damage and have resistance to cold damage.
    - Any creature that moves within
    5 feet of you for the first time on a turn or ends its turn there takes 1d10
    fire damage.
    - You can use your action to create a line of fire 15 feet long and
    5 feet wide extending from you in a direc- tion you choose. Each creature in
    the line must make a Dexterity saving throw. A creature takes 4d8 fire damage on
    a failed save, or half as much damage on a successful one.
    """

    name = "Investiture Of Flame"
    level = 6
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Druid", "Sorcerer", "Warlock", "Wizard")


class InvestitureOfIce(Spell):
    """Until the spell ends, ice rimes your body, and you gain the
    following benefits:

    - You are immune to cold damage and have resistance to fire
      damage.
    - You can move across difficult terrain created by ice or snow
      without spending extra movement.
    - The ground in a 10-foot radius around you is icy and is
      difficult terrain for creatures other than you. The radius moves
      with you.
    - You can use your action to create a 15-foot cone of freezing
      wind extending from your outstretched hand in a direction you
      choose. Each creature in the cone must make a Constitution
      saving throw. A creature takes 4d6 cold damage on a failed save,
      or half as much damage on a successful one. A creature that
      fails its save against this effect has its speed halved until
      the start of your next turn.

    """

    name = "Investiture Of Ice"
    level = 6
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Druid", "Sorcerer", "Warlock", "Wizard")


class InvestitureOfStone(Spell):
    """Until the spell ends, bits of rock spread across your body, and you gain the
    following benefits:

    - You have resistance to bludgeoning, piercing, and slashing
      damage from nonmagical weapons.
    - You can use your action to create a small earthquake on the
      ground in a 15-foot radius centered on you. Other creatures on
      that ground must succeed on a Dexterity saving throw or be
      knocked prone.
    - You can move across difficult terrain made of earth or stone
      without spending extra movement. You can move through solid
      earth or stone as if it was air and without destabilizing it,
      but you can't end your movement there. If you do so, you are
      ejected to the nearest unoccupied space, this spell ends, and
      you are stunned until the end of your next turn.

    """

    name = "Investiture Of Stone"
    level = 6
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Druid", "Sorcerer", "Warlock", "Wizard")


class InvestitureOfWind(Spell):
    """Until the spell ends, wind whirls around you, and you gain the following
    benefits:
    - Ranged weapon attacks made against you have disad- vantage on the
    attack roll.
    - You gain a flying speed of 60 feet. If you are still flying when
    the spell ends, you fall, unless you can some- how prevent it.
    - You can use
    your action to create a 15-foot cube of swirling wind centered on a point you
    can see within 60 feet of you. Each creature in that area must make a
    Constitution saving throw. A creature takes 2d10 bludgeoning damage on a failed
    save, or half as much damage on a successful one. If a Large or smaller creature
    fails the save, that creature is also pushed up to 10 feet away from the center
    of the cube.
    """

    name = "Investiture Of Wind"
    level = 6
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S")
    materials = ""
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Transmutation"
    classes = ("Druid", "Sorcerer", "Warlock", "Wizard")


class Invisibility(Spell):
    """A creature you touch becomes invisible until the spell ends. Anything the target
    is wearing or carrying is invisible as long as it is on the target's person.
    The spell ends for a target that attacks or casts a spell.

    At Higher Levels:
    When you cast this spell using a spell slot of 3rd level or higher, you can
    target one additional creature for each slot level above 2nd.
    """

    name = "Invisibility"
    level = 2
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "An eyelash encased in gum arabic"
    duration = "Concentration, up to 1 hour"
    ritual = False
    magic_school = "Illusion"
    classes = ("Bard", "Sorcerer", "Warlock", "Wizard")


class Invulnerability(Spell):
    """You are immune to all damage until the spell ends."""

    name = "Invulnerability"
    level = 9
    casting_time = "1 action"
    casting_range = "Self"
    components = ("V", "S", "M")
    materials = (
        "A small piece of adamantine worth at least 500 gp, which the spell consumes"
    )
    duration = "Concentration, up to 10 minutes"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Wizard",)



from dungeonsheets.spells.spells import Spell


class NegativeEnergyFlood(Spell):
    """You send ribbons of negative energy at one creature you can see within range.
    Unless the target is undead, it must make a Constitution saving throw, taking
    5d12 necrotic damage on a failed save, or half as much damage on a successful
    one. A target killed by this damage rises up as a zombie at the start of your
    next turn. The zombie pursues whatever creature it can see that is closest to
    it. Statistics for the zombie are in the Monster Manual. If you target an undead
    with this spell, the target doesn't make a saving throw. Instead, roll 5d12.
    The target gains half the total as temporary hit points.
    """

    name = "Negative Energy Flood"
    level = 5
    casting_time = "1 action"
    casting_range = "60 feet"
    components = ("",)
    materials = "A broken bone and a square of black silk"
    duration = "Instantaneous"
    ritual = False
    magic_school = "Necromancy"
    classes = ("Warlock", "Wizard")


class Nondetection(Spell):
    """For the duration, you hide a target that you touch from divination magic.
    The
    target can be a willing creature or a place or an object no larger than 10 feet
    in any dimension. The target can't be targeted by any divination magic or
    perceived through magical scrying sensors.
    """

    name = "Nondetection"
    level = 3
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = (
        "A pinch of diamond dust worth 25 gp sprinkled over the target, which the spell"
        " consumes"
    )
    duration = "8 hours"
    ritual = False
    magic_school = "Abjuration"
    classes = ("Bard", "Ranger", "Wizard")


class NystulsMagicAura(Spell):
    """You place an illusion on a creature or an object you touch so that divination
    spells reveal false information about it.
    The target can be a willing creature
    or an object that isn't being carried or worn by another creature.

    When you
    cast the spell, choose one or both of the following effects. The effect lasts
    for the duration. If you cast this spell on the same creature or object every
    day for 30 days, placing the same effect on it each time, the illusion lasts
    until it is dispelled.

    False Aura
    You change the way the target appears to
    spells and magical effects, such as detect magic, that detect magical auras. You
    can make a nonmagical object appear magical, a magical object appear
    nonmagical, or change the object's magical aura so that it appears to belong to
    a specific school of magic that you choose. When you use this effect on an
    object, you can make the false magic apparent to any creature that handles the
    item.

    Mask
    You change the way the target appears to spells and magical effects
    that detect creature types, such as a paladin's Divine Sense or the trigger of a
    sym bol spell. You choose a creature type and other spells and magical effects
    treat the target as if it were a creature of that type or of that alignment.
    """

    name = "Nystuls Magic Aura"
    level = 2
    casting_time = "1 action"
    casting_range = "Touch"
    components = ("V", "S", "M")
    materials = "A small square of silk"
    duration = "24 hours"
    ritual = False
    magic_school = "Illusion"
    classes = ("Wizard",)
